# -*- coding: utf-8 -*-
'''
    :codeauthor: Erik Johnson <erik@saltstack.com>
'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import logging
import os
import textwrap

# Import Salt Testing Libs
try:
    import pytest
except ImportError as import_error:
    pytest = None

from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase, skipIf
from tests.support.mock import (
    MagicMock,
    patch,
    mock_open,
    NO_MOCK,
    NO_MOCK_REASON
)

# Import Salt Libs
import salt.utils.files
import salt.utils.platform
import salt.utils.path
import salt.grains.core as core

# Import 3rd-party libs
from salt.ext import six
from salt._compat import ipaddress

log = logging.getLogger(__name__)

# Globals
IPv4Address = ipaddress.IPv4Address
IPv6Address = ipaddress.IPv6Address
IP4_LOCAL = '127.0.0.1'
IP4_ADD1 = '10.0.0.1'
IP4_ADD2 = '10.0.0.2'
IP6_LOCAL = '::1'
IP6_ADD1 = '2001:4860:4860::8844'
IP6_ADD2 = '2001:4860:4860::8888'
IP6_ADD_SCOPE = 'fe80::6238:e0ff:fe06:3f6b%enp2s0'
OS_RELEASE_DIR = os.path.join(os.path.dirname(__file__), "os-releases")
SOLARIS_DIR = os.path.join(os.path.dirname(__file__), 'solaris')


@skipIf(NO_MOCK, NO_MOCK_REASON)
@skipIf(not pytest, False)
class CoreGrainsTestCase(TestCase, LoaderModuleMockMixin):
    '''
    Test cases for core grains
    '''
    def setup_loader_modules(self):
        return {core: {}}

    @patch("os.path.isfile")
    def test_parse_etc_os_release(self, path_isfile_mock):
        path_isfile_mock.side_effect = lambda x: x == "/usr/lib/os-release"
        with salt.utils.files.fopen(os.path.join(OS_RELEASE_DIR, "ubuntu-17.10")) as os_release_file:
            os_release_content = os_release_file.read()
        with patch("salt.utils.files.fopen", mock_open(read_data=os_release_content)):
            os_release = core._parse_os_release(
                '/etc/os-release',
                '/usr/lib/os-release')
        self.assertEqual(os_release, {
            "NAME": "Ubuntu",
            "VERSION": "17.10 (Artful Aardvark)",
            "ID": "ubuntu",
            "ID_LIKE": "debian",
            "PRETTY_NAME": "Ubuntu 17.10",
            "VERSION_ID": "17.10",
            "HOME_URL": "https://www.ubuntu.com/",
            "SUPPORT_URL": "https://help.ubuntu.com/",
            "BUG_REPORT_URL": "https://bugs.launchpad.net/ubuntu/",
            "PRIVACY_POLICY_URL": "https://www.ubuntu.com/legal/terms-and-policies/privacy-policy",
            "VERSION_CODENAME": "artful",
            "UBUNTU_CODENAME": "artful",
        })

    def test_parse_cpe_name_wfn(self):
        '''
        Parse correct CPE_NAME data WFN formatted
        :return:
        '''
        for cpe, cpe_ret in [('cpe:/o:opensuse:leap:15.0',
                              {'phase': None, 'version': '15.0', 'product': 'leap',
                               'vendor': 'opensuse', 'part': 'operating system'}),
                             ('cpe:/o:vendor:product:42:beta',
                              {'phase': 'beta', 'version': '42', 'product': 'product',
                               'vendor': 'vendor', 'part': 'operating system'})]:
            ret = core._parse_cpe_name(cpe)
            for key in cpe_ret:
                assert key in ret
                assert cpe_ret[key] == ret[key]

    def test_parse_cpe_name_v23(self):
        '''
        Parse correct CPE_NAME data v2.3 formatted
        :return:
        '''
        for cpe, cpe_ret in [('cpe:2.3:o:microsoft:windows_xp:5.1.601:beta:*:*:*:*:*:*',
                              {'phase': 'beta', 'version': '5.1.601', 'product': 'windows_xp',
                               'vendor': 'microsoft', 'part': 'operating system'}),
                             ('cpe:2.3:h:corellian:millenium_falcon:1.0:*:*:*:*:*:*:*',
                              {'phase': None, 'version': '1.0', 'product': 'millenium_falcon',
                               'vendor': 'corellian', 'part': 'hardware'}),
                             ('cpe:2.3:*:dark_empire:light_saber:3.0:beta:*:*:*:*:*:*',
                              {'phase': 'beta', 'version': '3.0', 'product': 'light_saber',
                               'vendor': 'dark_empire', 'part': None})]:
            ret = core._parse_cpe_name(cpe)
            for key in cpe_ret:
                assert key in ret
                assert cpe_ret[key] == ret[key]

    def test_parse_cpe_name_broken(self):
        '''
        Parse broken CPE_NAME data
        :return:
        '''
        for cpe in ['cpe:broken', 'cpe:broken:in:all:ways:*:*:*:*',
                    'cpe:x:still:broken:123', 'who:/knows:what:is:here']:
            assert core._parse_cpe_name(cpe) == {}

    def test_missing_os_release(self):
        with patch('salt.utils.files.fopen', mock_open(read_data={})):
            os_release = core._parse_os_release('/etc/os-release', '/usr/lib/os-release')
        self.assertEqual(os_release, {})

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_gnu_slash_linux_in_os_name(self):
        '''
        Test to return a list of all enabled services
        '''
        _path_exists_map = {
            '/proc/1/cmdline': False
        }
        _path_isfile_map = {}
        _cmd_run_map = {
            'dpkg --print-architecture': 'amd64',
        }

        path_exists_mock = MagicMock(side_effect=lambda x: _path_exists_map[x])
        path_isfile_mock = MagicMock(
            side_effect=lambda x: _path_isfile_map.get(x, False)
        )
        cmd_run_mock = MagicMock(
            side_effect=lambda x: _cmd_run_map[x]
        )
        empty_mock = MagicMock(return_value={})

        orig_import = __import__
        if six.PY2:
            built_in = '__builtin__'
        else:
            built_in = 'builtins'

        def _import_mock(name, *args):
            if name == 'lsb_release':
                raise ImportError('No module named lsb_release')
            return orig_import(name, *args)

        # - Skip the first if statement
        # - Skip the selinux/systemd stuff (not pertinent)
        # - Skip the init grain compilation (not pertinent)
        # - Ensure that lsb_release fails to import
        # - Skip all the /etc/*-release stuff (not pertinent)
        # - Mock linux_distribution to give us the OS name that we want
        # - Make a bunch of functions return empty dicts, we don't care about
        #   these grains for the purposes of this test.
        # - Mock the osarch
        distro_mock = MagicMock(return_value=('Debian GNU/Linux', '8.3', ''))
        with patch.object(salt.utils.platform, 'is_proxy',
                          MagicMock(return_value=False)), \
                patch.object(core, '_linux_bin_exists',
                             MagicMock(return_value=False)), \
                patch.object(os.path, 'exists', path_exists_mock), \
                patch('{0}.__import__'.format(built_in), side_effect=_import_mock), \
                patch.object(os.path, 'isfile', path_isfile_mock), \
                patch.object(core, '_parse_lsb_release', empty_mock), \
                patch.object(core, '_parse_os_release', empty_mock), \
                patch.object(core, '_parse_lsb_release', empty_mock), \
                patch.object(core, 'linux_distribution', distro_mock), \
                patch.object(core, '_linux_cpudata', empty_mock), \
                patch.object(core, '_linux_gpu_data', empty_mock), \
                patch.object(core, '_memdata', empty_mock), \
                patch.object(core, '_hw_data', empty_mock), \
                patch.object(core, '_virtual', empty_mock), \
                patch.object(core, '_ps', empty_mock), \
                patch.dict(core.__salt__, {'cmd.run': cmd_run_mock}):
            os_grains = core.os_data()

        self.assertEqual(os_grains.get('os_family'), 'Debian')

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_suse_os_from_cpe_data(self):
        '''
        Test if 'os' grain is parsed from CPE_NAME of /etc/os-release
        '''
        _path_exists_map = {
            '/proc/1/cmdline': False
        }
        _os_release_map = {
            'NAME': 'SLES',
            'VERSION': '12-SP1',
            'VERSION_ID': '12.1',
            'PRETTY_NAME': 'SUSE Linux Enterprise Server 12 SP1',
            'ID': 'sles',
            'ANSI_COLOR': '0;32',
            'CPE_NAME': 'cpe:/o:suse:sles:12:sp1'
        }

        path_exists_mock = MagicMock(side_effect=lambda x: _path_exists_map[x])
        empty_mock = MagicMock(return_value={})
        osarch_mock = MagicMock(return_value="amd64")
        os_release_mock = MagicMock(return_value=_os_release_map)

        orig_import = __import__
        if six.PY2:
            built_in = '__builtin__'
        else:
            built_in = 'builtins'

        def _import_mock(name, *args):
            if name == 'lsb_release':
                raise ImportError('No module named lsb_release')
            return orig_import(name, *args)

        distro_mock = MagicMock(
            return_value=('SUSE Linux Enterprise Server ', '12', 'x86_64')
        )

        # - Skip the first if statement
        # - Skip the selinux/systemd stuff (not pertinent)
        # - Skip the init grain compilation (not pertinent)
        # - Ensure that lsb_release fails to import
        # - Skip all the /etc/*-release stuff (not pertinent)
        # - Mock linux_distribution to give us the OS name that we want
        # - Mock the osarch
        with patch.object(salt.utils.platform, 'is_proxy',
                          MagicMock(return_value=False)), \
                patch.object(core, '_linux_bin_exists',
                             MagicMock(return_value=False)), \
                patch.object(os.path, 'exists', path_exists_mock), \
                patch('{0}.__import__'.format(built_in),
                      side_effect=_import_mock), \
                patch.object(os.path, 'isfile', MagicMock(return_value=False)), \
                patch.object(core, '_parse_os_release', os_release_mock), \
                patch.object(core, '_parse_lsb_release', empty_mock), \
                patch.object(core, 'linux_distribution', distro_mock), \
                patch.object(core, '_linux_gpu_data', empty_mock), \
                patch.object(core, '_hw_data', empty_mock), \
                patch.object(core, '_linux_cpudata', empty_mock), \
                patch.object(core, '_virtual', empty_mock), \
                patch.dict(core.__salt__, {'cmd.run': osarch_mock}):
            os_grains = core.os_data()

        self.assertEqual(os_grains.get('os_family'), 'Suse')
        self.assertEqual(os_grains.get('os'), 'SUSE')

    def _run_os_grains_tests(self, os_release_filename, os_release_map, expectation):
        path_isfile_mock = MagicMock(side_effect=lambda x: x in os_release_map.get('files', []))
        empty_mock = MagicMock(return_value={})
        osarch_mock = MagicMock(return_value="amd64")
        if os_release_filename:
            os_release_data = core._parse_os_release(
                os.path.join(OS_RELEASE_DIR, os_release_filename)
            )
        else:
            os_release_data = os_release_map.get('os_release_file', {})
        os_release_mock = MagicMock(return_value=os_release_data)

        orig_import = __import__
        if six.PY2:
            built_in = '__builtin__'
        else:
            built_in = 'builtins'

        def _import_mock(name, *args):
            if name == 'lsb_release':
                raise ImportError('No module named lsb_release')
            return orig_import(name, *args)

        suse_release_file = os_release_map.get('suse_release_file')

        file_contents = {'/proc/1/cmdline': ''}
        if suse_release_file:
            file_contents['/etc/SuSE-release'] = suse_release_file

        # - Skip the first if statement
        # - Skip the selinux/systemd stuff (not pertinent)
        # - Skip the init grain compilation (not pertinent)
        # - Ensure that lsb_release fails to import
        # - Skip all the /etc/*-release stuff (not pertinent)
        # - Mock linux_distribution to give us the OS name that we want
        # - Mock the osarch
        distro_mock = MagicMock(return_value=os_release_map['linux_distribution'])
        with patch.object(salt.utils.platform, 'is_proxy', MagicMock(return_value=False)), \
                patch.object(core, '_linux_bin_exists', MagicMock(return_value=False)), \
                patch.object(os.path, 'exists', path_isfile_mock), \
                patch('{0}.__import__'.format(built_in), side_effect=_import_mock), \
                patch.object(os.path, 'isfile', path_isfile_mock), \
                patch.object(core, '_parse_os_release', os_release_mock), \
                patch.object(core, '_parse_lsb_release', empty_mock), \
                patch('salt.utils.files.fopen', mock_open(read_data=file_contents)), \
                patch.object(core, 'linux_distribution', distro_mock), \
                patch.object(core, '_linux_gpu_data', empty_mock), \
                patch.object(core, '_linux_cpudata', empty_mock), \
                patch.object(core, '_virtual', empty_mock), \
                patch.dict(core.__salt__, {'cmd.run': osarch_mock}):
            os_grains = core.os_data()

        grains = {k: v for k, v in os_grains.items()
                  if k in set(["os", "os_family", "osfullname", "oscodename", "osfinger",
                               "osrelease", "osrelease_info", "osmajorrelease"])}
        self.assertEqual(grains, expectation)

    def _run_suse_os_grains_tests(self, os_release_map, expectation):
        os_release_map['linux_distribution'] = ('SUSE test', 'version', 'arch')
        expectation['os'] = 'SUSE'
        expectation['os_family'] = 'Suse'
        self._run_os_grains_tests(None, os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_suse_os_grains_sles11sp3(self):
        '''
        Test if OS grains are parsed correctly in SLES 11 SP3
        '''
        _os_release_map = {
            'suse_release_file': textwrap.dedent('''
                SUSE Linux Enterprise Server 11 (x86_64)
                VERSION = 11
                PATCHLEVEL = 3
                '''),
            'files': ["/etc/SuSE-release"],
        }
        expectation = {
            'oscodename': 'SUSE Linux Enterprise Server 11 SP3',
            'osfullname': "SLES",
            'osrelease': '11.3',
            'osrelease_info': (11, 3),
            'osmajorrelease': 11,
            'osfinger': 'SLES-11',
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_suse_os_grains_sles11sp4(self):
        '''
        Test if OS grains are parsed correctly in SLES 11 SP4
        '''
        _os_release_map = {
            'os_release_file': {
                'NAME': 'SLES',
                'VERSION': '11.4',
                'VERSION_ID': '11.4',
                'PRETTY_NAME': 'SUSE Linux Enterprise Server 11 SP4',
                'ID': 'sles',
                'ANSI_COLOR': '0;32',
                'CPE_NAME': 'cpe:/o:suse:sles:11:4'
            },
        }
        expectation = {
            'oscodename': 'SUSE Linux Enterprise Server 11 SP4',
            'osfullname': "SLES",
            'osrelease': '11.4',
            'osrelease_info': (11, 4),
            'osmajorrelease': 11,
            'osfinger': 'SLES-11',
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_suse_os_grains_sles12(self):
        '''
        Test if OS grains are parsed correctly in SLES 12
        '''
        _os_release_map = {
            'os_release_file': {
                'NAME': 'SLES',
                'VERSION': '12',
                'VERSION_ID': '12',
                'PRETTY_NAME': 'SUSE Linux Enterprise Server 12',
                'ID': 'sles',
                'ANSI_COLOR': '0;32',
                'CPE_NAME': 'cpe:/o:suse:sles:12'
            },
        }
        expectation = {
            'oscodename': 'SUSE Linux Enterprise Server 12',
            'osfullname': "SLES",
            'osrelease': '12',
            'osrelease_info': (12,),
            'osmajorrelease': 12,
            'osfinger': 'SLES-12',
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_suse_os_grains_sles12sp1(self):
        '''
        Test if OS grains are parsed correctly in SLES 12 SP1
        '''
        _os_release_map = {
            'os_release_file': {
                'NAME': 'SLES',
                'VERSION': '12-SP1',
                'VERSION_ID': '12.1',
                'PRETTY_NAME': 'SUSE Linux Enterprise Server 12 SP1',
                'ID': 'sles',
                'ANSI_COLOR': '0;32',
                'CPE_NAME': 'cpe:/o:suse:sles:12:sp1'
            },
        }
        expectation = {
            'oscodename': 'SUSE Linux Enterprise Server 12 SP1',
            'osfullname': "SLES",
            'osrelease': '12.1',
            'osrelease_info': (12, 1),
            'osmajorrelease': 12,
            'osfinger': 'SLES-12',
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_suse_os_grains_opensuse_leap_42_1(self):
        '''
        Test if OS grains are parsed correctly in openSUSE Leap 42.1
        '''
        _os_release_map = {
            'os_release_file': {
                'NAME': 'openSUSE Leap',
                'VERSION': '42.1',
                'VERSION_ID': '42.1',
                'PRETTY_NAME': 'openSUSE Leap 42.1 (x86_64)',
                'ID': 'opensuse',
                'ANSI_COLOR': '0;32',
                'CPE_NAME': 'cpe:/o:opensuse:opensuse:42.1'
            },
        }
        expectation = {
            'oscodename': 'openSUSE Leap 42.1 (x86_64)',
            'osfullname': "Leap",
            'osrelease': '42.1',
            'osrelease_info': (42, 1),
            'osmajorrelease': 42,
            'osfinger': 'Leap-42',
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_suse_os_grains_tumbleweed(self):
        '''
        Test if OS grains are parsed correctly in openSUSE Tumbleweed
        '''
        _os_release_map = {
            'os_release_file': {
                'NAME': 'openSUSE',
                'VERSION': 'Tumbleweed',
                'VERSION_ID': '20160504',
                'PRETTY_NAME': 'openSUSE Tumbleweed (20160504) (x86_64)',
                'ID': 'opensuse',
                'ANSI_COLOR': '0;32',
                'CPE_NAME': 'cpe:/o:opensuse:opensuse:20160504'
            },
        }
        expectation = {
            'oscodename': 'openSUSE Tumbleweed (20160504) (x86_64)',
            'osfullname': "Tumbleweed",
            'osrelease': '20160504',
            'osrelease_info': (20160504,),
            'osmajorrelease': 20160504,
            'osfinger': 'Tumbleweed-20160504',
        }
        self._run_suse_os_grains_tests(_os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_debian_7_os_grains(self):
        '''
        Test if OS grains are parsed correctly in Debian 7 "wheezy"
        '''
        _os_release_map = {
            'linux_distribution': ('debian', '7.11', ''),
        }
        expectation = {
            'os': 'Debian',
            'os_family': 'Debian',
            'oscodename': 'wheezy',
            'osfullname': 'Debian GNU/Linux',
            'osrelease': '7',
            'osrelease_info': (7,),
            'osmajorrelease': 7,
            'osfinger': 'Debian-7',
        }
        self._run_os_grains_tests("debian-7", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_debian_8_os_grains(self):
        '''
        Test if OS grains are parsed correctly in Debian 8 "jessie"
        '''
        _os_release_map = {
            'linux_distribution': ('debian', '8.10', ''),
        }
        expectation = {
            'os': 'Debian',
            'os_family': 'Debian',
            'oscodename': 'jessie',
            'osfullname': 'Debian GNU/Linux',
            'osrelease': '8',
            'osrelease_info': (8,),
            'osmajorrelease': 8,
            'osfinger': 'Debian-8',
        }
        self._run_os_grains_tests("debian-8", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_debian_9_os_grains(self):
        '''
        Test if OS grains are parsed correctly in Debian 9 "stretch"
        '''
        _os_release_map = {
            'linux_distribution': ('debian', '9.3', ''),
        }
        expectation = {
            'os': 'Debian',
            'os_family': 'Debian',
            'oscodename': 'stretch',
            'osfullname': 'Debian GNU/Linux',
            'osrelease': '9',
            'osrelease_info': (9,),
            'osmajorrelease': 9,
            'osfinger': 'Debian-9',
        }
        self._run_os_grains_tests("debian-9", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_ubuntu_xenial_os_grains(self):
        '''
        Test if OS grains are parsed correctly in Ubuntu 16.04 "Xenial Xerus"
        '''
        _os_release_map = {
            'linux_distribution': ('Ubuntu', '16.04', 'xenial'),
        }
        expectation = {
            'os': 'Ubuntu',
            'os_family': 'Debian',
            'oscodename': 'xenial',
            'osfullname': 'Ubuntu',
            'osrelease': '16.04',
            'osrelease_info': (16, 4),
            'osmajorrelease': 16,
            'osfinger': 'Ubuntu-16.04',
        }
        self._run_os_grains_tests("ubuntu-16.04", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_ubuntu_artful_os_grains(self):
        '''
        Test if OS grains are parsed correctly in Ubuntu 17.10 "Artful Aardvark"
        '''
        _os_release_map = {
            'linux_distribution': ('Ubuntu', '17.10', 'artful'),
        }
        expectation = {
            'os': 'Ubuntu',
            'os_family': 'Debian',
            'oscodename': 'artful',
            'osfullname': 'Ubuntu',
            'osrelease': '17.10',
            'osrelease_info': (17, 10),
            'osmajorrelease': 17,
            'osfinger': 'Ubuntu-17.10',
        }
        self._run_os_grains_tests("ubuntu-17.10", _os_release_map, expectation)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_linux_memdata(self):
        '''
        Test memdata on Linux systems
        '''
        _proc_meminfo = textwrap.dedent('''\
            MemTotal:       16277028 kB
            SwapTotal:       4789244 kB''')
        with patch('salt.utils.files.fopen', mock_open(read_data=_proc_meminfo)):
            memdata = core._linux_memdata()
        self.assertEqual(memdata.get('mem_total'), 15895)
        self.assertEqual(memdata.get('swap_total'), 4676)

    @skipIf(salt.utils.platform.is_windows(), 'System is Windows')
    def test_bsd_memdata(self):
        '''
        Test to memdata on *BSD systems
        '''
        _path_exists_map = {}
        _path_isfile_map = {}
        _cmd_run_map = {
            'freebsd-version -u': '10.3-RELEASE',
            '/sbin/sysctl -n hw.physmem': '2121781248',
            '/sbin/sysctl -n vm.swap_total': '419430400'
        }

        path_exists_mock = MagicMock(side_effect=lambda x: _path_exists_map[x])
        path_isfile_mock = MagicMock(
            side_effect=lambda x: _path_isfile_map.get(x, False)
        )
        cmd_run_mock = MagicMock(
            side_effect=lambda x: _cmd_run_map[x]
        )
        empty_mock = MagicMock(return_value={})

        mock_freebsd_uname = ('FreeBSD',
                              'freebsd10.3-hostname-8148',
                              '10.3-RELEASE',
                              'FreeBSD 10.3-RELEASE #0 r297264: Fri Mar 25 02:10:02 UTC 2016     root@releng1.nyi.freebsd.org:/usr/obj/usr/src/sys/GENERIC',
                              'amd64',
                              'amd64')

        with patch('platform.uname',
                   MagicMock(return_value=mock_freebsd_uname)):
            with patch.object(salt.utils.platform, 'is_linux',
                              MagicMock(return_value=False)):
                with patch.object(salt.utils.platform, 'is_freebsd',
                                  MagicMock(return_value=True)):
                    # Skip the first if statement
                    with patch.object(salt.utils.platform, 'is_proxy',
                                      MagicMock(return_value=False)):
                        # Skip the init grain compilation (not pertinent)
                        with patch.object(os.path, 'exists', path_exists_mock):
                            with patch('salt.utils.path.which') as mock:
                                mock.return_value = '/sbin/sysctl'
                                # Make a bunch of functions return empty dicts,
                                # we don't care about these grains for the
                                # purposes of this test.
                                with patch.object(
                                        core,
                                        '_bsd_cpudata',
                                        empty_mock):
                                    with patch.object(
                                            core,
                                            '_hw_data',
                                            empty_mock):
                                        with patch.object(
                                                core,
                                                '_virtual',
                                                empty_mock):
                                            with patch.object(
                                                    core,
                                                    '_ps',
                                                    empty_mock):
                                                # Mock the osarch
                                                with patch.dict(
                                                        core.__salt__,
                                                        {'cmd.run': cmd_run_mock}):
                                                    os_grains = core.os_data()

        self.assertEqual(os_grains.get('mem_total'), 2023)
        self.assertEqual(os_grains.get('swap_total'), 400)

    @skipIf(salt.utils.platform.is_windows(), 'System is Windows')
    def test_docker_virtual(self):
        '''
        Test if OS grains are parsed correctly in Ubuntu Xenial Xerus
        '''
        with patch.object(os.path, 'isdir', MagicMock(return_value=False)):
            with patch.object(os.path,
                              'isfile',
                              MagicMock(side_effect=lambda x: True if x == '/proc/1/cgroup' else False)):
                for cgroup_substr in (':/system.slice/docker', ':/docker/',
                                       ':/docker-ce/'):
                    cgroup_data = \
                        '10:memory{0}a_long_sha256sum'.format(cgroup_substr)
                    log.debug(
                        'Testing Docker cgroup substring \'%s\'', cgroup_substr)
                    with patch('salt.utils.files.fopen', mock_open(read_data=cgroup_data)):
                        with patch.dict(core.__salt__, {'cmd.run_all': MagicMock()}):
                            self.assertEqual(
                                core._virtual({'kernel': 'Linux'}).get('virtual_subtype'),
                                'Docker'
                            )

    @skipIf(salt.utils.platform.is_windows(), 'System is Windows')
    def test_xen_virtual(self):
        '''
        Test if OS grains are parsed correctly in Ubuntu Xenial Xerus
        '''
        with patch.object(os.path, 'isfile', MagicMock(return_value=False)):
            with patch.dict(core.__salt__, {'cmd.run': MagicMock(return_value='')}), \
                patch.object(os.path,
                             'isfile',
                             MagicMock(side_effect=lambda x: True if x == '/sys/bus/xen/drivers/xenconsole' else False)):
                log.debug('Testing Xen')
                self.assertEqual(
                    core._virtual({'kernel': 'Linux'}).get('virtual_subtype'),
                    'Xen PV DomU'
                )

    def _check_ipaddress(self, value, ip_v):
        '''
        check if ip address in a list is valid
        '''
        for val in value:
            assert isinstance(val, six.string_types)
            ip_method = 'is_ipv{0}'.format(ip_v)
            self.assertTrue(getattr(salt.utils.network, ip_method)(val))

    def _check_empty(self, key, value, empty):
        '''
        if empty is False and value does not exist assert error
        if empty is True and value exists assert error
        '''
        if not empty and not value:
            raise Exception("{0} is empty, expecting a value".format(key))
        elif empty and value:
            raise Exception("{0} is suppose to be empty. value: {1} \
                            exists".format(key, value))

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_fqdn_return(self):
        '''
        test ip4 and ip6 return values
        '''
        net_ip4_mock = [IP4_LOCAL, IP4_ADD1, IP4_ADD2]
        net_ip6_mock = [IP6_LOCAL, IP6_ADD1, IP6_ADD2]

        self._run_fqdn_tests(net_ip4_mock, net_ip6_mock,
                             ip4_empty=False, ip6_empty=False)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_fqdn6_empty(self):
        '''
        test when ip6 is empty
        '''
        net_ip4_mock = [IP4_LOCAL, IP4_ADD1, IP4_ADD2]
        net_ip6_mock = []

        self._run_fqdn_tests(net_ip4_mock, net_ip6_mock,
                             ip4_empty=False)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_fqdn4_empty(self):
        '''
        test when ip4 is empty
        '''
        net_ip4_mock = []
        net_ip6_mock = [IP6_LOCAL, IP6_ADD1, IP6_ADD2]

        self._run_fqdn_tests(net_ip4_mock, net_ip6_mock,
                             ip6_empty=False)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    def test_fqdn_all_empty(self):
        '''
        test when both ip4 and ip6 are empty
        '''
        net_ip4_mock = []
        net_ip6_mock = []

        self._run_fqdn_tests(net_ip4_mock, net_ip6_mock)

    def _run_fqdn_tests(self, net_ip4_mock, net_ip6_mock,
                        ip6_empty=True, ip4_empty=True):

        def _check_type(key, value, ip4_empty, ip6_empty):
            '''
            check type and other checks
            '''
            assert isinstance(value, list)

            if '4' in key:
                self._check_empty(key, value, ip4_empty)
                self._check_ipaddress(value, ip_v='4')
            elif '6' in key:
                self._check_empty(key, value, ip6_empty)
                self._check_ipaddress(value, ip_v='6')

        ip4_mock = [(2, 1, 6, '', (IP4_ADD1, 0)),
                    (2, 3, 0, '', (IP4_ADD2, 0))]
        ip6_mock = [(10, 1, 6, '', (IP6_ADD1, 0, 0, 0)),
                    (10, 3, 0, '', (IP6_ADD2, 0, 0, 0))]

        with patch.dict(core.__opts__, {'ipv6': False}):
            with patch.object(salt.utils.network, 'ip_addrs',
                             MagicMock(return_value=net_ip4_mock)):
                with patch.object(salt.utils.network, 'ip_addrs6',
                                 MagicMock(return_value=net_ip6_mock)):
                    with patch.object(core.socket, 'getaddrinfo', side_effect=[ip4_mock, ip6_mock]):
                        get_fqdn = core.ip_fqdn()
                        ret_keys = ['fqdn_ip4', 'fqdn_ip6', 'ipv4', 'ipv6']
                        for key in ret_keys:
                            value = get_fqdn[key]
                            _check_type(key, value, ip4_empty, ip6_empty)

    @skipIf(not salt.utils.platform.is_linux(), 'System is not Linux')
    @patch.object(salt.utils.platform, 'is_windows', MagicMock(return_value=False))
    @patch('salt.grains.core.__opts__', {'ipv6': False})
    def test_dns_return(self):
        '''
        test the return for a dns grain. test for issue:
        https://github.com/saltstack/salt/issues/41230
        '''
        resolv_mock = {'domain': '', 'sortlist': [], 'nameservers':
                   [ipaddress.IPv4Address(IP4_ADD1),
                    ipaddress.IPv6Address(IP6_ADD1),
                    IP6_ADD_SCOPE], 'ip4_nameservers':
                   [ipaddress.IPv4Address(IP4_ADD1)],
                   'search': ['test.saltstack.com'], 'ip6_nameservers':
                   [ipaddress.IPv6Address(IP6_ADD1),
                    IP6_ADD_SCOPE], 'options': []}
        ret = {'dns': {'domain': '', 'sortlist': [], 'nameservers':
                       [IP4_ADD1, IP6_ADD1,
                        IP6_ADD_SCOPE], 'ip4_nameservers':
                       [IP4_ADD1], 'search': ['test.saltstack.com'],
                       'ip6_nameservers': [IP6_ADD1, IP6_ADD_SCOPE],
                       'options': []}}
        with patch.object(salt.utils.dns, 'parse_resolv', MagicMock(return_value=resolv_mock)):
            assert core.dns() == ret

    def test_core_virtual(self):
        '''
        test virtual grain with cmd virt-what
        '''
        virt = 'kvm'
        with patch.object(salt.utils.platform, 'is_windows',
                          MagicMock(return_value=False)):
            with patch.object(salt.utils.path, 'which',
                              MagicMock(return_value=True)):
                with patch.dict(core.__salt__, {'cmd.run_all':
                                                MagicMock(return_value={'pid': 78,
                                                                        'retcode': 0,
                                                                        'stderr': '',
                                                                        'stdout': virt})}):
                    osdata = {'kernel': 'test', }
                    ret = core._virtual(osdata)
                    self.assertEqual(ret['virtual'], virt)

    def test_solaris_sparc_s7zone(self):
        '''
        verify productname grain for s7 zone
        '''
        expectation = {
                'productname': 'SPARC S7-2',
                'product': 'SPARC S7-2',
        }
        with salt.utils.files.fopen(os.path.join(SOLARIS_DIR, 'prtconf.s7-zone')) as sparc_return_data:
            this_sparc_return_data = '\n'.join(sparc_return_data.readlines())
            this_sparc_return_data += '\n'
        self._check_solaris_sparc_productname_grains(this_sparc_return_data, expectation)

    def test_solaris_sparc_s7(self):
        '''
        verify productname grain for s7
        '''
        expectation = {
                'productname': 'SPARC S7-2',
                'product': 'SPARC S7-2',
        }
        with salt.utils.files.fopen(os.path.join(SOLARIS_DIR, 'prtdiag.s7')) as sparc_return_data:
            this_sparc_return_data = '\n'.join(sparc_return_data.readlines())
            this_sparc_return_data += '\n'
        self._check_solaris_sparc_productname_grains(this_sparc_return_data, expectation)

    def test_solaris_sparc_t5220(self):
        '''
        verify productname grain for t5220
        '''
        expectation = {
                'productname': 'SPARC Enterprise T5220',
                'product': 'SPARC Enterprise T5220',
        }
        with salt.utils.files.fopen(os.path.join(SOLARIS_DIR, 'prtdiag.t5220')) as sparc_return_data:
            this_sparc_return_data = '\n'.join(sparc_return_data.readlines())
            this_sparc_return_data += '\n'
        self._check_solaris_sparc_productname_grains(this_sparc_return_data, expectation)

    def test_solaris_sparc_t5220zone(self):
        '''
        verify productname grain for t5220 zone
        '''
        expectation = {
                'productname': 'SPARC Enterprise T5220',
                'product': 'SPARC Enterprise T5220',
        }
        with salt.utils.files.fopen(os.path.join(SOLARIS_DIR, 'prtconf.t5220-zone')) as sparc_return_data:
            this_sparc_return_data = '\n'.join(sparc_return_data.readlines())
            this_sparc_return_data += '\n'
        self._check_solaris_sparc_productname_grains(this_sparc_return_data, expectation)

    def _check_solaris_sparc_productname_grains(self, prtdata, expectation):
        '''
        verify product grains on solaris sparc
        '''
        import platform
        path_isfile_mock = MagicMock(side_effect=lambda x: x in ['/etc/release'])
        with salt.utils.files.fopen(os.path.join(OS_RELEASE_DIR, "solaris-11.3")) as os_release_file:
            os_release_content = os_release_file.readlines()
        uname_mock = MagicMock(return_value=(
            'SunOS', 'testsystem', '5.11', '11.3', 'sunv4', 'sparc'
        ))
        with patch.object(platform, 'uname', uname_mock), \
                patch.object(salt.utils.platform, 'is_proxy',
                             MagicMock(return_value=False)), \
                patch.object(salt.utils.platform, 'is_linux',
                             MagicMock(return_value=False)), \
                patch.object(salt.utils.platform, 'is_windows',
                             MagicMock(return_value=False)), \
                patch.object(salt.utils.platform, 'is_smartos',
                             MagicMock(return_value=False)), \
                patch.object(salt.utils.path, 'which_bin',
                             MagicMock(return_value=None)), \
                patch.object(os.path, 'isfile', path_isfile_mock), \
                patch('salt.utils.files.fopen',
                      mock_open(read_data=os_release_content)) as os_release_file, \
                patch.object(core, '_sunos_cpudata',
                             MagicMock(return_value={
                                 'cpuarch': 'sparcv9',
                                 'num_cpus': '1',
                                 'cpu_model': 'MOCK_CPU_MODEL',
                                 'cpu_flags': []})), \
                patch.object(core, '_memdata',
                             MagicMock(return_value={'mem_total': 16384})), \
                patch.object(core, '_virtual',
                             MagicMock(return_value={})), \
                patch.object(core, '_ps', MagicMock(return_value={})), \
                patch.object(salt.utils.path, 'which',
                             MagicMock(return_value=True)), \
                patch.dict(core.__salt__,
                           {'cmd.run': MagicMock(return_value=prtdata)}):
            os_grains = core.os_data()
        grains = {k: v for k, v in os_grains.items()
                  if k in set(['product', 'productname'])}
        self.assertEqual(grains, expectation)

    @patch('os.path.isfile')
    @patch('os.path.isdir')
    def test_core_virtual_unicode(self, mock_file, mock_dir):
        '''
        test virtual grain with unicode character in product_name file
        '''
        def path_side_effect(path):
            if path == '/sys/devices/virtual/dmi/id/product_name':
                return True
            return False

        virt = 'kvm'
        mock_file.side_effect = path_side_effect
        mock_dir.side_effect = path_side_effect
        with patch.object(salt.utils.platform, 'is_windows',
                          MagicMock(return_value=False)):
            with patch.object(salt.utils.path, 'which',
                              MagicMock(return_value=True)):
                with patch.dict(core.__salt__, {'cmd.run_all':
                                                MagicMock(return_value={'pid': 78,
                                                                        'retcode': 0,
                                                                        'stderr': '',
                                                                        'stdout': virt})}):
                    with patch('salt.utils.files.fopen',
                               mock_open(read_data='嗨')):
                        osdata = {'kernel': 'Linux', }
                        ret = core._virtual(osdata)
                        self.assertEqual(ret['virtual'], virt)

    @patch('salt.utils.path.which', MagicMock(return_value='/usr/sbin/sysctl'))
    def test_osx_memdata_with_comma(self):
        '''
        test osx memdata method when comma returns
        '''
        def _cmd_side_effect(cmd):
            if 'hw.memsize' in cmd:
                return '4294967296'
            elif 'vm.swapusage' in cmd:
                return 'total = 1024,00M  used = 160,75M  free = 863,25M  (encrypted)'
        with patch.dict(core.__salt__, {'cmd.run': MagicMock(side_effect=_cmd_side_effect)}):
            ret = core._osx_memdata()
            assert ret['swap_total'] == 1024
            assert ret['mem_total'] == 4096

    @patch('salt.utils.path.which', MagicMock(return_value='/usr/sbin/sysctl'))
    def test_osx_memdata(self):
        '''
        test osx memdata
        '''
        def _cmd_side_effect(cmd):
            if 'hw.memsize' in cmd:
                return '4294967296'
            elif 'vm.swapusage' in cmd:
                return 'total = 0.00M  used = 0.00M  free = 0.00M  (encrypted)'
        with patch.dict(core.__salt__, {'cmd.run': MagicMock(side_effect=_cmd_side_effect)}):
            ret = core._osx_memdata()
            assert ret['swap_total'] == 0
            assert ret['mem_total'] == 4096



# -*- coding: utf-8 -*-
'''
Tests for salt.utils.yamlencoding
'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
import salt.utils.yaml
import salt.utils.yamlencoding
from tests.support.unit import TestCase


class YamlEncodingTestCase(TestCase):

    def test_yaml_dquote(self):
        for teststr in (r'"\ []{}"',):
            self.assertEqual(teststr, salt.utils.yaml.safe_load(salt.utils.yamlencoding.yaml_dquote(teststr)))

    def test_yaml_dquote_doesNotAddNewLines(self):
        teststr = '"' * 100
        self.assertNotIn('\n', salt.utils.yamlencoding.yaml_dquote(teststr))

    def test_yaml_squote(self):
        ret = salt.utils.yamlencoding.yaml_squote(r'"')
        self.assertEqual(ret, r"""'"'""")

    def test_yaml_squote_doesNotAddNewLines(self):
        teststr = "'" * 100
        self.assertNotIn('\n', salt.utils.yamlencoding.yaml_squote(teststr))

    def test_yaml_encode(self):
        for testobj in (None, True, False, '[7, 5]', '"monkey"', 5, 7.5, "2014-06-02 15:30:29.7"):
            self.assertEqual(testobj, salt.utils.yaml.safe_load(salt.utils.yamlencoding.yaml_encode(testobj)))

        for testobj in ({}, [], set()):
            self.assertRaises(TypeError, salt.utils.yamlencoding.yaml_encode, testobj)



# -*- coding: utf-8 -*-
'''
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
'''

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals
import logging

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase, skipIf
from tests.support.mock import (
    Mock,
    MagicMock,
    patch,
    NO_MOCK,
    NO_MOCK_REASON
)

# Import Salt Libs
import salt.states.user as user

log = logging.getLogger(__name__)


@skipIf(NO_MOCK, NO_MOCK_REASON)
class UserTestCase(TestCase, LoaderModuleMockMixin):
    '''
        Validate the user state
    '''
    def setup_loader_modules(self):
        return {user: {}}

    def test_present(self):
        '''
            Test to ensure that the named user is present with
            the specified properties
        '''
        ret = {'name': 'salt',
               'changes': {},
               'result': False,
               'comment': ''}
        mock_false = MagicMock(return_value=False)
        mock_empty_list = MagicMock(return_value=[])
        with patch.dict(user.__grains__, {"kernel": 'Linux'}):
            with patch.dict(user.__salt__, {'group.info': mock_false,
                                            'user.info': mock_empty_list,
                                            "user.chkey": mock_empty_list,
                                            'user.add': mock_false}):
                ret.update({'comment': 'The following group(s) are'
                            ' not present: salt'})
                self.assertDictEqual(user.present('salt', groups=['salt']), ret)

                mock_false = MagicMock(side_effect=[{'key': 'value'}, {'key': 'value'},
                                              {'key': 'value'}, False, False])
                with patch.object(user, '_changes', mock_false):
                    with patch.dict(user.__opts__, {"test": True}):
                        ret.update(
                            {'comment': 'The following user attributes are set '
                                        'to be changed:\n'
                                        'key: value\n',
                             'result': None})
                        self.assertDictEqual(user.present('salt'), ret)

                    with patch.dict(user.__opts__, {"test": False}):
                        # pylint: disable=repr-flag-used-in-string
                        comment = (
                            'These values could not be changed: {0!r}'
                            .format({'key': 'value'})
                        )
                        # pylint: enable=repr-flag-used-in-string
                        ret.update({'comment': comment, 'result': False})
                        self.assertDictEqual(user.present('salt'), ret)

                        with patch.dict(user.__opts__, {"test": True}):
                            ret.update({'comment': 'User salt set to'
                                        ' be added', 'result': None})
                            self.assertDictEqual(user.present('salt'), ret)

                        with patch.dict(user.__opts__, {"test": False}):
                            ret.update({'comment': 'Failed to create new'
                                        ' user salt', 'result': False})
                            self.assertDictEqual(user.present('salt'), ret)

    def test_present_invalid_uid_change(self):
        mock_info = MagicMock(side_effect=[
            {'uid': 5000,
             'gid': 5000,
             'groups': ['foo'],
             'home': '/home/foo',
             'fullname': 'Foo Bar'}
        ])
        dunder_salt = {'user.info': mock_info,
                       'file.group_to_gid': MagicMock(side_effect=['foo']),
                       'file.gid_to_group': MagicMock(side_effect=[5000])}
        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {'kernel': 'Linux'}), \
                patch.dict(user.__salt__, dunder_salt):
            ret = user.present('foo', uid=5001)
            # State should have failed
            self.assertFalse(ret['result'])
            # Only one of uid/gid should have been flagged in the comment
            self.assertEqual(ret['comment'].count('not permitted'), 1)

    def test_present_invalid_gid_change(self):
        mock_info = MagicMock(side_effect=[
            {'uid': 5000,
             'gid': 5000,
             'groups': ['foo'],
             'home': '/home/foo',
             'fullname': 'Foo Bar'}
        ])
        dunder_salt = {'user.info': mock_info,
                       'file.group_to_gid': MagicMock(side_effect=['foo']),
                       'file.gid_to_group': MagicMock(side_effect=[5000])}
        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {'kernel': 'Linux'}), \
                patch.dict(user.__salt__, dunder_salt):
            ret = user.present('foo', gid=5001)
            # State should have failed
            self.assertFalse(ret['result'])
            # Only one of uid/gid should have been flagged in the comment
            self.assertEqual(ret['comment'].count('not permitted'), 1)

    def test_present_invalid_uid_gid_change(self):
        mock_info = MagicMock(side_effect=[
            {'uid': 5000,
             'gid': 5000,
             'groups': ['foo'],
             'home': '/home/foo',
             'fullname': 'Foo Bar'}
        ])
        dunder_salt = {'user.info': mock_info,
                       'file.group_to_gid': MagicMock(side_effect=['foo']),
                       'file.gid_to_group': MagicMock(side_effect=[5000])}
        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {'kernel': 'Linux'}), \
                patch.dict(user.__salt__, dunder_salt):
            ret = user.present('foo', uid=5001, gid=5001)
            # State should have failed
            self.assertFalse(ret['result'])
            # Both the uid and gid should have been flagged in the comment
            self.assertEqual(ret['comment'].count('not permitted'), 2)

    def test_present_uid_gid_change(self):
        before = {'uid': 5000,
                  'gid': 5000,
                  'groups': ['foo'],
                  'home': '/home/foo',
                  'fullname': 'Foo Bar'}
        after = {'uid': 5001,
                 'gid': 5001,
                 'groups': ['othergroup'],
                 'home': '/home/foo',
                 'fullname': 'Foo Bar'}
        # user.info should be called 4 times. Once the first time that
        # _changes() is called, once before and after changes are applied (to
        # get the before/after for the changes dict, and one last time to
        # confirm that no changes still need to be made.
        mock_info = MagicMock(side_effect=[before, before, after, after])
        mock_group_to_gid = MagicMock(side_effect=['foo', 'othergroup'])
        mock_gid_to_group = MagicMock(side_effect=[5000, 5001])
        dunder_salt = {'user.info': mock_info,
                       'user.chuid': Mock(),
                       'user.chgid': Mock(),
                       'file.group_to_gid': mock_group_to_gid,
                       'file.gid_to_group': mock_gid_to_group}
        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {'kernel': 'Linux'}), \
                patch.dict(user.__salt__, dunder_salt), \
                patch.dict(user.__opts__, {'test': False}), \
                patch('os.path.isdir', MagicMock(return_value=True)):
            ret = user.present(
                'foo',
                uid=5001,
                gid=5001,
                allow_uid_change=True,
                allow_gid_change=True)
            self.assertEqual(
                ret,
                {'comment': 'Updated user foo',
                 'changes': {'gid': 5001,
                             'uid': 5001,
                             'groups': ['othergroup']},
                 'name': 'foo',
                 'result': True}
            )

    def test_absent(self):
        '''
            Test to ensure that the named user is absent
        '''
        ret = {'name': 'salt',
               'changes': {},
               'result': None,
               'comment': ''}
        mock = MagicMock(side_effect=[True, True, False])
        mock1 = MagicMock(return_value=False)
        with patch.dict(user.__salt__, {'user.info': mock,
                                        'user.delete': mock1,
                                        'group.info': mock1}):
            with patch.dict(user.__opts__, {"test": True}):
                ret.update({'comment': 'User salt set for removal'})
                self.assertDictEqual(user.absent('salt'), ret)

            with patch.dict(user.__opts__, {"test": False}):
                ret.update({'comment': 'Failed to remove user salt',
                            'result': False})
                self.assertDictEqual(user.absent('salt'), ret)

            ret.update({'comment': 'User salt is not present',
                        'result': True})
            self.assertDictEqual(user.absent('salt'), ret)

    def test_changes(self):
        '''
        Test salt.states.user._changes
        '''
        mock_info = MagicMock(
            return_value={
                'uid': 5000,
                 'gid': 5000,
                 'groups': ['foo'],
                 'home': '/home/foo',
                 'fullname': 'Foo Bar',
            }
        )
        shadow_info = MagicMock(
            return_value={
                'min': 2,
                'max': 88888,
                'inact': 77,
                'warn': 14,
            }
        )
        shadow_hash = MagicMock(return_value='abcd')
        dunder_salt = {'user.info': mock_info,
                       'shadow.info': shadow_info,
                       'shadow.default_hash': shadow_hash,
                       'file.group_to_gid': MagicMock(side_effect=['foo']),
                       'file.gid_to_group': MagicMock(side_effect=[5000])}

        def mock_exists(*args):
            return True

        # side_effect used because these mocks should only be called once
        with patch.dict(user.__grains__, {'kernel': 'Linux'}), \
                patch.dict(user.__salt__, dunder_salt), \
                patch.dict(user.__opts__, {"test": False}), \
                patch('os.path.isdir', mock_exists):
            ret = user._changes('foo', maxdays=999999, inactdays=0, warndays=7)
            assert ret == {
                'maxdays': 999999,
                'mindays': 0,
                'fullname': '',
                'warndays': 7,
                'inactdays': 0
            }



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Alexandru Bleotu <alexandru.bleotu@morganstanley.com>`

    Tests for functions in salt.modules.esxcluster
'''

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.modules.esxcluster as esxcluster

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase, skipIf
from tests.support.mock import (
    MagicMock,
    patch,
    NO_MOCK,
    NO_MOCK_REASON
)


@skipIf(NO_MOCK, NO_MOCK_REASON)
class GetDetailsTestCase(TestCase, LoaderModuleMockMixin):
    '''Tests for salt.modules.esxcluster.get_details'''
    def setup_loader_modules(self):
        return {esxcluster: {'__virtual__':
                             MagicMock(return_value='esxcluster'),
                             '__proxy__': {}}}

    def test_get_details(self):
        mock_get_details = MagicMock()
        with patch.dict(esxcluster.__proxy__,
                        {'esxcluster.get_details': mock_get_details}):
            esxcluster.get_details()
        mock_get_details.assert_called_once_with()



# -*- coding: utf-8 -*-
'''

    integration.reactor.reactor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test Salt's reactor system
'''

# Import Python libs
from __future__ import absolute_import

# Import Salt testing libs
from tests.support.case import ModuleCase
from tests.support.helpers import flaky
from tests.support.mixins import SaltMinionEventAssertsMixin

# Import Salt libs
import salt.utils.event


class ReactorTest(ModuleCase, SaltMinionEventAssertsMixin):
    '''
    Test Salt's reactor system
    '''

    @flaky
    def test_ping_reaction(self):
        '''
        Fire an event on the master and ensure
        that it pings the minion
        '''
        # Create event bus connection
        e = salt.utils.event.get_event('minion', sock_dir=self.minion_opts['sock_dir'], opts=self.minion_opts)

        e.fire_event({'a': 'b'}, '/test_event')

        self.assertMinionEventReceived({'a': 'b'})



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import string
import random

# Import Salt Testing libs
from tests.support.case import ModuleCase
from tests.support.unit import skipIf
from tests.support.helpers import (
    destructiveTest,
    skip_if_not_root,
    requires_system_grains
)

# Import Salt libs
import salt.utils.platform

# Import 3rd-party libs
from salt.ext.six.moves import range  # pylint: disable=import-error,redefined-builtin


@destructiveTest
@skipIf(not salt.utils.platform.is_linux(), 'These tests can only be run on linux')
@skip_if_not_root
class UseraddModuleTestLinux(ModuleCase):

    def setUp(self):
        super(UseraddModuleTestLinux, self).setUp()
        os_grain = self.run_function('grains.item', ['kernel'])
        if os_grain['kernel'] not in ('Linux', 'Darwin'):
            self.skipTest(
                'Test not applicable to \'{kernel}\' kernel'.format(
                    **os_grain
                )
            )

    def __random_string(self, size=6):
        return 'RS-' + ''.join(
            random.choice(string.ascii_uppercase + string.digits)
            for x in range(size)
        )

    @requires_system_grains
    def test_groups_includes_primary(self, grains):
        # Let's create a user, which usually creates the group matching the
        # name
        uname = self.__random_string()
        if self.run_function('user.add', [uname]) is not True:
            # Skip because creating is not what we're testing here
            self.run_function('user.delete', [uname, True, True])
            self.skipTest('Failed to create user')

        try:
            uinfo = self.run_function('user.info', [uname])
            if grains['os_family'] in ('Suse',):
                self.assertIn('users', uinfo['groups'])
            else:
                self.assertIn(uname, uinfo['groups'])

            # This uid is available, store it
            uid = uinfo['uid']

            self.run_function('user.delete', [uname, True, True])

            # Now, a weird group id
            gname = self.__random_string()
            if self.run_function('group.add', [gname]) is not True:
                self.run_function('group.delete', [gname, True, True])
                self.skipTest('Failed to create group')

            ginfo = self.run_function('group.info', [gname])

            # And create the user with that gid
            if self.run_function('user.add', [uname, uid, ginfo['gid']]) is False:
                # Skip because creating is not what we're testing here
                self.run_function('user.delete', [uname, True, True])
                self.skipTest('Failed to create user')

            uinfo = self.run_function('user.info', [uname])
            self.assertIn(gname, uinfo['groups'])

        except AssertionError:
            self.run_function('user.delete', [uname, True, True])
            raise

    def test_user_primary_group(self):
        '''
        Tests the primary_group function
        '''
        name = 'saltyuser'

        # Create a user to test primary group function
        if self.run_function('user.add', [name]) is not True:
            self.run_function('user.delete', [name])
            self.skipTest('Failed to create a user')

        try:
            # Test useradd.primary_group
            primary_group = self.run_function('user.primary_group', [name])
            uid_info = self.run_function('user.info', [name])
            self.assertIn(primary_group, uid_info['groups'])

        except Exception:
            self.run_function('user.delete', [name])
            raise


@destructiveTest
@skipIf(not salt.utils.platform.is_windows(), 'These tests can only be run on Windows')
@skip_if_not_root
class UseraddModuleTestWindows(ModuleCase):

    def __random_string(self, size=6):
        return 'RS-' + ''.join(
            random.choice(string.ascii_uppercase + string.digits)
            for x in range(size))

    def setUp(self):
        self.user_name = self.__random_string()
        self.group_name = self.__random_string()

    def tearDown(self):
        self.run_function('user.delete', [self.user_name, True, True])
        self.run_function('group.delete', [self.group_name])

    def _add_user(self):
        '''
        helper class to add user
        '''
        if self.run_function('user.add', [self.user_name]) is False:
            self.run_function('user.delete', [self.user_name, True, True])
            self.skipTest('Failed to create user')

    def _add_group(self):
        '''
        helper class to add group
        '''
        if self.run_function('group.add', [self.group_name]) is False:
            # Skip because creating is not what we're testing here
            self.run_function('group.delete', [self.group_name, True, True])
            self.skipTest('Failed to create group')

    def test_add_user(self):
        '''
        Test adding a user
        '''
        user_name = self.__random_string()

        try:
            if self.run_function('user.add', [user_name]) is False:
                self.run_function('user.delete', [user_name, True, True])
                self.skipTest('Failed to create user')

            user_list = self.run_function('user.list_users')
            self.assertIn(user_name, user_list)

        except AssertionError:
            raise

        finally:
            self.run_function('user.delete', [user_name, True, True])

    def test_add_group(self):
        '''
        Test adding a user
        '''
        group_name = self.__random_string()
        try:
            if self.run_function('group.add', [group_name]) is False:
                # Skip because creating is not what we're testing here
                self.run_function('group.delete', [group_name, True, True])
                self.skipTest('Failed to create group')

            group_list = self.run_function('group.list_groups')
            self.assertIn(group_name, group_list)

        except AssertionError:
            raise

        finally:
            self.run_function('group.delete', [group_name])

    def test_add_user_to_group(self):
        '''
        Test adding a user to a group
        '''
        group_name = self.__random_string()
        user_name = self.__random_string()
        try:
            # Let's create a group
            if self.run_function(
                    'group.add', [group_name])['result'] is not True:
                self.run_function('group.delete', [group_name, True, True])
                self.skipTest('Failed to create group')

            # And create the user as a member of that group
            if self.run_function(
                    'user.add', [user_name], groups=group_name) is False:
                self.run_function('user.delete', [user_name, True, True])
                self.skipTest('Failed to create user')

            user_info = self.run_function('user.info', [user_name])
            self.assertIn(group_name, user_info['groups'])

        except AssertionError:
            raise

        finally:
            self.run_function('user.delete', [user_name, True, True])
            self.run_function('group.delete', [group_name])

    def test_add_user_addgroup(self):
        '''
        Test adding a user to a group with groupadd
        '''
        self._add_group()
        self._add_user()
        self.run_function('user.addgroup', [self.user_name, self.group_name])
        info = self.run_function('user.info', [self.user_name])
        self.assertEqual(info['groups'], [self.group_name])

    def test_user_chhome(self):
        '''
        Test changing a users home dir
        '''
        self._add_user()
        user_dir = r'c:\salt'
        self.run_function('user.chhome', [self.user_name, user_dir])
        info = self.run_function('user.info', [self.user_name])
        self.assertEqual(info['home'], user_dir)

    def test_user_chprofile(self):
        '''
        Test changing a users profile
        '''
        self._add_user()
        config = r'c:\salt\config'
        self.run_function('user.chprofile', [self.user_name, config])
        info = self.run_function('user.info', [self.user_name])
        self.assertEqual(info['profile'], config)

    def test_user_chfullname(self):
        '''
        Test changing a users fullname
        '''
        self._add_user()
        name = 'Salt Test'
        self.run_function('user.chfullname', [self.user_name, name])
        info = self.run_function('user.info', [self.user_name])
        self.assertEqual(info['fullname'], name)

    def test_user_delete(self):
        '''
        Test deleting a user
        '''
        self._add_user()
        self.assertTrue(self.run_function('user.info', [self.user_name])['active'])
        self.run_function('user.delete', [self.user_name])
        self.assertEqual({}, self.run_function('user.info', [self.user_name]))

    def test_user_removegroup(self):
        '''
        Test removing a group
        '''
        self._add_user()
        self._add_group()
        self.run_function('user.addgroup', [self.user_name, self.group_name])
        self.assertIn(self.group_name, self.run_function('user.list_groups', [self.user_name]))
        self.run_function('user.removegroup', [self.user_name, self.group_name])
        self.assertNotIn(self.group_name, self.run_function('user.list_groups', [self.user_name]))

    def test_user_rename(self):
        '''
        Test changing a users name
        '''
        self._add_user()
        name = 'newuser'
        self.run_function('user.rename', [self.user_name, name])
        info = self.run_function('user.info', [name])
        self.assertTrue(info['active'])

        #delete new user
        self.run_function('user.delete', [name, True, True])

    def test_user_setpassword(self):
        '''
        Test setting a password
        '''
        self._add_user()
        passwd = 'sup3rs3cr3T!'
        self.assertTrue(self.run_function('user.setpassword', [self.user_name, passwd]))



# -*- coding: utf-8 -*-
# Import Python libs
from __future__ import absolute_import, unicode_literals, print_function
import logging
import socket
import textwrap

# Import Salt Testing libs
from tests.support.unit import skipIf
from tests.support.unit import TestCase
from tests.support.mock import (
    MagicMock,
    mock_open,
    patch,
    NO_MOCK,
    NO_MOCK_REASON,
)

# Import salt libs
import salt.utils.network as network
from salt._compat import ipaddress

log = logging.getLogger(__name__)

LINUX = '''\
eth0      Link encap:Ethernet  HWaddr e0:3f:49:85:6a:af
          inet addr:10.10.10.56  Bcast:10.10.10.255  Mask:255.255.252.0
          inet6 addr: fe80::e23f:49ff:fe85:6aaf/64 Scope:Link
          UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
          RX packets:643363 errors:0 dropped:0 overruns:0 frame:0
          TX packets:196539 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:1000
          RX bytes:386388355 (368.4 MiB)  TX bytes:25600939 (24.4 MiB)

lo        Link encap:Local Loopback
          inet addr:127.0.0.1  Mask:255.0.0.0
          inet6 addr: ::1/128 Scope:Host
          UP LOOPBACK RUNNING  MTU:65536  Metric:1
          RX packets:548901 errors:0 dropped:0 overruns:0 frame:0
          TX packets:548901 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:0
          RX bytes:613479895 (585.0 MiB)  TX bytes:613479895 (585.0 MiB)
'''

FREEBSD = '''
em0: flags=8843<UP,BROADCAST,RUNNING,SIMPLEX,MULTICAST> metric 0 mtu 1500
        options=4219b<RXCSUM,TXCSUM,VLAN_MTU,VLAN_HWTAGGING,VLAN_HWCSUM,TSO4,WOL_MAGIC,VLAN_HWTSO>
        ether 00:30:48:ff:ff:ff
        inet 10.10.10.250 netmask 0xffffffe0 broadcast 10.10.10.255
        inet 10.10.10.56 netmask 0xffffffc0 broadcast 10.10.10.63
        media: Ethernet autoselect (1000baseT <full-duplex>)
        status: active
em1: flags=8c02<BROADCAST,OACTIVE,SIMPLEX,MULTICAST> metric 0 mtu 1500
        options=4219b<RXCSUM,TXCSUM,VLAN_MTU,VLAN_HWTAGGING,VLAN_HWCSUM,TSO4,WOL_MAGIC,VLAN_HWTSO>
        ether 00:30:48:aa:aa:aa
        media: Ethernet autoselect
        status: no carrier
plip0: flags=8810<POINTOPOINT,SIMPLEX,MULTICAST> metric 0 mtu 1500
lo0: flags=8049<UP,LOOPBACK,RUNNING,MULTICAST> metric 0 mtu 16384
        options=3<RXCSUM,TXCSUM>
        inet6 fe80::1%lo0 prefixlen 64 scopeid 0x8
        inet6 ::1 prefixlen 128
        inet 127.0.0.1 netmask 0xff000000
        nd6 options=3<PERFORMNUD,ACCEPT_RTADV>
tun0: flags=8051<UP,POINTOPOINT,RUNNING,MULTICAST> metric 0 mtu 1500
        options=80000<LINKSTATE>
        inet 10.12.0.1 --> 10.12.0.2 netmask 0xffffffff
        Opened by PID 1964
'''

SOLARIS = '''\
lo0: flags=2001000849<UP,LOOPBACK,RUNNING,MULTICAST,IPv4,VIRTUAL> mtu 8232 index 1
        inet 127.0.0.1 netmask ff000000
net0: flags=100001100943<UP,BROADCAST,RUNNING,PROMISC,MULTICAST,ROUTER,IPv4,PHYSRUNNING> mtu 1500 index 2
        inet 10.10.10.38 netmask ffffffe0 broadcast 10.10.10.63
ilbint0: flags=110001100843<UP,BROADCAST,RUNNING,MULTICAST,ROUTER,IPv4,VRRP,PHYSRUNNING> mtu 1500 index 3
        inet 10.6.0.11 netmask ffffff00 broadcast 10.6.0.255
ilbext0: flags=110001100843<UP,BROADCAST,RUNNING,MULTICAST,ROUTER,IPv4,VRRP,PHYSRUNNING> mtu 1500 index 4
        inet 10.10.11.11 netmask ffffffe0 broadcast 10.10.11.31
ilbext0:1: flags=110001100843<UP,BROADCAST,RUNNING,MULTICAST,ROUTER,IPv4,VRRP,PHYSRUNNING> mtu 1500 index 4
        inet 10.10.11.12 netmask ffffffe0 broadcast 10.10.11.31
vpn0: flags=1000011008d1<UP,POINTOPOINT,RUNNING,NOARP,MULTICAST,ROUTER,IPv4,PHYSRUNNING> mtu 1480 index 5
        inet tunnel src 10.10.11.12 tunnel dst 10.10.5.5
        tunnel hop limit 64
        inet 10.6.0.14 --> 10.6.0.15 netmask ff000000
lo0: flags=2002000849<UP,LOOPBACK,RUNNING,MULTICAST,IPv6,VIRTUAL> mtu 8252 index 1
        inet6 ::1/128
net0: flags=120002004941<UP,RUNNING,PROMISC,MULTICAST,DHCP,IPv6,PHYSRUNNING> mtu 1500 index 2
        inet6 fe80::221:9bff:fefd:2a22/10
ilbint0: flags=120002000840<RUNNING,MULTICAST,IPv6,PHYSRUNNING> mtu 1500 index 3
        inet6 ::/0
ilbext0: flags=120002000840<RUNNING,MULTICAST,IPv6,PHYSRUNNING> mtu 1500 index 4
        inet6 ::/0
vpn0: flags=120002200850<POINTOPOINT,RUNNING,MULTICAST,NONUD,IPv6,PHYSRUNNING> mtu 1480 index 5
        inet tunnel src 10.10.11.12 tunnel dst 10.10.5.5
        tunnel hop limit 64
        inet6 ::/0 --> fe80::b2d6:7c10
'''

NETBSD = '''\
vioif0: flags=0x8943<UP,BROADCAST,RUNNING,PROMISC,SIMPLEX,MULTICAST> mtu 1500
        ec_capabilities=1<VLAN_MTU>
        ec_enabled=0
        address: 00:a0:98:e6:83:18
        inet 192.168.1.80/24 broadcast 192.168.1.255 flags 0x0
        inet6 fe80::2a0:98ff:fee6:8318%vioif0/64 flags 0x0 scopeid 0x1
lo0: flags=0x8049<UP,LOOPBACK,RUNNING,MULTICAST> mtu 33624
        inet 127.0.0.1/8 flags 0x0
        inet6 ::1/128 flags 0x20<NODAD>
        inet6 fe80::1%lo0/64 flags 0x0 scopeid 0x2
'''

FREEBSD_SOCKSTAT = '''\
USER    COMMAND     PID     FD  PROTO  LOCAL ADDRESS    FOREIGN ADDRESS
root    python2.7   1294    41  tcp4   127.0.0.1:61115  127.0.0.1:4506
'''

FREEBSD_SOCKSTAT_WITH_FAT_PID = '''\
USER     COMMAND    PID   FD PROTO  LOCAL ADDRESS    FOREIGN ADDRESS
salt-master python2.781106 35 tcp4  127.0.0.1:61115  127.0.0.1:4506
'''

IPV4_SUBNETS = {True: ('10.10.0.0/24',),
                False: ('10.10.0.0', '10.10.0.0/33', 'FOO', 9, '0.9.800.1000/24')}
IPV6_SUBNETS = {True: ('::1/128',),
                False: ('::1', '::1/129', 'FOO', 9, 'aj01::feac/64')}


@skipIf(NO_MOCK, NO_MOCK_REASON)
class NetworkTestCase(TestCase):

    def test_sanitize_host(self):
        ret = network.sanitize_host('10.1./2.$3')
        self.assertEqual(ret, '10.1.2.3')

    def test_host_to_ips(self):
        '''
        NOTE: When this test fails it's usually because the IP address has
        changed. In these cases, we just need to update the IP address in the
        assertion.
        '''
        def _side_effect(host, *args):
            try:
                return {
                    'github.com': [
                        (2, 1, 6, '', ('192.30.255.112', 0)),
                        (2, 1, 6, '', ('192.30.255.113', 0)),
                    ],
                    'ipv6host.foo': [
                        (socket.AF_INET6, 1, 6, '', ('2001:a71::1', 0, 0, 0)),
                    ],
                }[host]
            except KeyError:
                raise socket.gaierror(-2, 'Name or service not known')

        getaddrinfo_mock = MagicMock(side_effect=_side_effect)
        with patch.object(socket, 'getaddrinfo', getaddrinfo_mock):
            # Test host that can be resolved
            ret = network.host_to_ips('github.com')
            self.assertEqual(ret, ['192.30.255.112', '192.30.255.113'])
            # Test ipv6
            ret = network.host_to_ips('ipv6host.foo')
            self.assertEqual(ret, ['2001:a71::1'])
            # Test host that can't be resolved
            ret = network.host_to_ips('someothersite.com')
            self.assertEqual(ret, None)

    def test_generate_minion_id(self):
        self.assertTrue(network.generate_minion_id())

    def test__generate_minion_id_with_unicode_in_etc_hosts(self):
        '''
        Test that unicode in /etc/hosts doesn't raise an error when
        _generate_minion_id() helper is called to gather the hosts.
        '''
        content = textwrap.dedent('''\
        # 以下为主机名解析
        ## ccc
        127.0.0.1       localhost thisismyhostname     # 本机
        ''')
        fopen_mock = mock_open(read_data={'/etc/hosts': content})
        with patch('salt.utils.files.fopen', fopen_mock):
            assert 'thisismyhostname' in network._generate_minion_id()

    def test_is_ip(self):
        self.assertTrue(network.is_ip('10.10.0.3'))
        self.assertFalse(network.is_ip('0.9.800.1000'))

    def test_is_ipv4(self):
        self.assertTrue(network.is_ipv4('10.10.0.3'))
        self.assertFalse(network.is_ipv4('10.100.1'))
        self.assertFalse(network.is_ipv4('2001:db8:0:1:1:1:1:1'))

    def test_is_ipv6(self):
        self.assertTrue(network.is_ipv6('2001:db8:0:1:1:1:1:1'))
        self.assertTrue(network.is_ipv6('0:0:0:0:0:0:0:1'))
        self.assertTrue(network.is_ipv6('::1'))
        self.assertTrue(network.is_ipv6('::'))
        self.assertTrue(network.is_ipv6('2001:0db8:85a3:0000:0000:8a2e:0370:7334'))
        self.assertTrue(network.is_ipv6('2001:0db8:85a3::8a2e:0370:7334'))
        self.assertFalse(network.is_ipv6('2001:0db8:0370:7334'))
        self.assertFalse(network.is_ipv6('2001:0db8:::0370:7334'))
        self.assertFalse(network.is_ipv6('10.0.1.2'))
        self.assertFalse(network.is_ipv6('2001.0db8.85a3.0000.0000.8a2e.0370.7334'))

    def test_parse_host_port(self):
        _ip = ipaddress.ip_address
        good_host_ports = {
            '10.10.0.3': (_ip('10.10.0.3').compressed, None),
            '10.10.0.3:1234': (_ip('10.10.0.3').compressed, 1234),
            '2001:0db8:85a3::8a2e:0370:7334': (_ip('2001:0db8:85a3::8a2e:0370:7334').compressed, None),
            '[2001:0db8:85a3::8a2e:0370:7334]:1234': (_ip('2001:0db8:85a3::8a2e:0370:7334').compressed, 1234),
            '2001:0db8:85a3::7334': (_ip('2001:0db8:85a3::7334').compressed, None),
            '[2001:0db8:85a3::7334]:1234': (_ip('2001:0db8:85a3::7334').compressed, 1234)
        }
        bad_host_ports = [
            '10.10.0.3/24',
            '10.10.0.3::1234',
            '2001:0db8:0370:7334',
            '2001:0db8:0370::7334]:1234',
            '2001:0db8:0370:0:a:b:c:d:1234'
        ]
        for host_port, assertion_value in good_host_ports.items():
            host = port = None
            host, port = network.parse_host_port(host_port)
            self.assertEqual((host, port), assertion_value)

        for host_port in bad_host_ports:
            try:
                self.assertRaises(ValueError, network.parse_host_port, host_port)
            except AssertionError as _e_:
                log.error('bad host_port value: "%s" failed to trigger ValueError exception', host_port)
                raise _e_

    def test_is_subnet(self):
        for subnet_data in (IPV4_SUBNETS, IPV6_SUBNETS):
            for item in subnet_data[True]:
                log.debug('Testing that %s is a valid subnet', item)
                self.assertTrue(network.is_subnet(item))
            for item in subnet_data[False]:
                log.debug('Testing that %s is not a valid subnet', item)
                self.assertFalse(network.is_subnet(item))

    def test_is_ipv4_subnet(self):
        for item in IPV4_SUBNETS[True]:
            log.debug('Testing that %s is a valid subnet', item)
            self.assertTrue(network.is_ipv4_subnet(item))
        for item in IPV4_SUBNETS[False]:
            log.debug('Testing that %s is not a valid subnet', item)
            self.assertFalse(network.is_ipv4_subnet(item))

    def test_is_ipv6_subnet(self):
        for item in IPV6_SUBNETS[True]:
            log.debug('Testing that %s is a valid subnet', item)
            self.assertTrue(network.is_ipv6_subnet(item))
        for item in IPV6_SUBNETS[False]:
            log.debug('Testing that %s is not a valid subnet', item)
            self.assertFalse(network.is_ipv6_subnet(item))

    def test_cidr_to_ipv4_netmask(self):
        self.assertEqual(network.cidr_to_ipv4_netmask(24), '255.255.255.0')
        self.assertEqual(network.cidr_to_ipv4_netmask(21), '255.255.248.0')
        self.assertEqual(network.cidr_to_ipv4_netmask(17), '255.255.128.0')
        self.assertEqual(network.cidr_to_ipv4_netmask(9), '255.128.0.0')
        self.assertEqual(network.cidr_to_ipv4_netmask(36), '')
        self.assertEqual(network.cidr_to_ipv4_netmask('lol'), '')

    def test_number_of_set_bits_to_ipv4_netmask(self):
        set_bits_to_netmask = network._number_of_set_bits_to_ipv4_netmask(0xffffff00)
        self.assertEqual(set_bits_to_netmask, '255.255.255.0')
        set_bits_to_netmask = network._number_of_set_bits_to_ipv4_netmask(0xffff6400)

    def test_hex2ip(self):
        self.assertEqual(network.hex2ip('0x4A7D2B63'), '74.125.43.99')
        self.assertEqual(network.hex2ip('0x4A7D2B63', invert=True), '99.43.125.74')

    def test_interfaces_ifconfig_linux(self):
        interfaces = network._interfaces_ifconfig(LINUX)
        self.assertEqual(interfaces,
                         {'eth0': {'hwaddr': 'e0:3f:49:85:6a:af',
                                   'inet': [{'address': '10.10.10.56',
                                             'broadcast': '10.10.10.255',
                                             'netmask': '255.255.252.0'}],
                                   'inet6': [{'address': 'fe80::e23f:49ff:fe85:6aaf',
                                              'prefixlen': '64',
                                              'scope': 'link'}],
                                   'up': True},
                          'lo': {'inet': [{'address': '127.0.0.1',
                                           'netmask': '255.0.0.0'}],
                                 'inet6': [{'address': '::1',
                                            'prefixlen': '128',
                                            'scope': 'host'}],
                                 'up': True}}
        )

    def test_interfaces_ifconfig_freebsd(self):
        interfaces = network._interfaces_ifconfig(FREEBSD)
        self.assertEqual(interfaces,
                         {'': {'up': False},
                          'em0': {'hwaddr': '00:30:48:ff:ff:ff',
                                  'inet': [{'address': '10.10.10.250',
                                            'broadcast': '10.10.10.255',
                                            'netmask': '255.255.255.224'},
                                           {'address': '10.10.10.56',
                                            'broadcast': '10.10.10.63',
                                            'netmask': '255.255.255.192'}],
                                  'up': True},
                          'em1': {'hwaddr': '00:30:48:aa:aa:aa',
                                  'up': False},
                          'lo0': {'inet': [{'address': '127.0.0.1',
                                            'netmask': '255.0.0.0'}],
                                  'inet6': [{'address': 'fe80::1',
                                             'prefixlen': '64',
                                             'scope': '0x8'},
                                            {'address': '::1',
                                             'prefixlen': '128',
                                             'scope': None}],
                                  'up': True},
                          'plip0': {'up': False},
                          'tun0': {'inet': [{'address': '10.12.0.1',
                                             'netmask': '255.255.255.255'}],
                                   'up': True}}

        )

    def test_interfaces_ifconfig_solaris(self):
        with patch('salt.utils.platform.is_sunos', lambda: True):
            interfaces = network._interfaces_ifconfig(SOLARIS)
            expected_interfaces = {'ilbint0':
                                       {'inet6': [],
                                        'inet': [{'broadcast': '10.6.0.255',
                                                  'netmask': '255.255.255.0',
                                                  'address': '10.6.0.11'}],
                                        'up': True},
                                   'lo0':
                                       {'inet6': [{'prefixlen': '128',
                                                   'address': '::1'}],
                                       'inet': [{'netmask': '255.0.0.0',
                                                 'address': '127.0.0.1'}],
                                        'up': True},
                                   'ilbext0': {'inet6': [],
                                               'inet': [{'broadcast': '10.10.11.31',
                                                         'netmask': '255.255.255.224',
                                                         'address': '10.10.11.11'},
                                                        {'broadcast': '10.10.11.31',
                                                         'netmask': '255.255.255.224',
                                                         'address': '10.10.11.12'}],
                                               'up': True},
                                   'vpn0': {'inet6': [],
                                            'inet': [{'netmask': '255.0.0.0',
                                                      'address': '10.6.0.14'}],
                                            'up': True},
                                   'net0': {'inet6': [{'prefixlen': '10',
                                                       'address': 'fe80::221:9bff:fefd:2a22'}],
                                   'inet': [{'broadcast': '10.10.10.63',
                                             'netmask': '255.255.255.224',
                                             'address': '10.10.10.38'}],
                                            'up': True}}
            self.assertEqual(interfaces, expected_interfaces)

    def test_interfaces_ifconfig_netbsd(self):
        interfaces = network._netbsd_interfaces_ifconfig(NETBSD)
        self.assertEqual(interfaces,
                          {'lo0': {'inet': [{'address': '127.0.0.1', 'netmask': '255.0.0.0'}],
                                   'inet6': [{'address': 'fe80::1',
                                               'prefixlen': '64',
                                               'scope': 'lo0'}],
                                   'up': True},
                           'vioif0': {'hwaddr': '00:a0:98:e6:83:18',
                                      'inet': [{'address': '192.168.1.80',
                                                 'broadcast': '192.168.1.255',
                                                 'netmask': '255.255.255.0'}],
                                      'inet6': [{'address': 'fe80::2a0:98ff:fee6:8318',
                                                  'prefixlen': '64',
                                                  'scope': 'vioif0'}],
                                      'up': True}}
        )

    def test_freebsd_remotes_on(self):
        with patch('salt.utils.platform.is_sunos', lambda: False):
            with patch('salt.utils.platform.is_freebsd', lambda: True):
                with patch('subprocess.check_output',
                           return_value=FREEBSD_SOCKSTAT):
                    remotes = network._freebsd_remotes_on('4506', 'remote')
                    self.assertEqual(remotes, set(['127.0.0.1']))

    def test_freebsd_remotes_on_with_fat_pid(self):
        with patch('salt.utils.platform.is_sunos', lambda: False):
            with patch('salt.utils.platform.is_freebsd', lambda: True):
                with patch('subprocess.check_output',
                           return_value=FREEBSD_SOCKSTAT_WITH_FAT_PID):
                    remotes = network._freebsd_remotes_on('4506', 'remote')
                    self.assertEqual(remotes, set(['127.0.0.1']))

    def test_generate_minion_id_distinct(self):
        '''
        Test if minion IDs are distinct in the pool.

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='nodename')), \
                patch('socket.gethostname', MagicMock(return_value='hostname')), \
                patch('socket.getfqdn', MagicMock(return_value='hostname.domainname.blank')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'attrname', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['1.2.3.4', '5.6.7.8'])):
            self.assertEqual(network._generate_minion_id(),
                             ['hostname.domainname.blank', 'nodename', 'hostname', '1.2.3.4', '5.6.7.8'])

    def test_generate_minion_id_127_name(self):
        '''
        Test if minion IDs can be named 127.foo

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='127')), \
                patch('socket.gethostname', MagicMock(return_value='127')), \
                patch('socket.getfqdn', MagicMock(return_value='127.domainname.blank')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'attrname', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['1.2.3.4', '5.6.7.8'])):
            self.assertEqual(network._generate_minion_id(),
                             ['127.domainname.blank', '127', '1.2.3.4', '5.6.7.8'])

    def test_generate_minion_id_127_name_startswith(self):
        '''
        Test if minion IDs can be named starting from "127"

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='127890')), \
                patch('socket.gethostname', MagicMock(return_value='127890')), \
                patch('socket.getfqdn', MagicMock(return_value='127890.domainname.blank')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'attrname', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['1.2.3.4', '5.6.7.8'])):
            self.assertEqual(network._generate_minion_id(),
                             ['127890.domainname.blank', '127890', '1.2.3.4', '5.6.7.8'])

    def test_generate_minion_id_duplicate(self):
        '''
        Test if IP addresses in the minion IDs are distinct in the pool

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='hostname')), \
                patch('socket.gethostname', MagicMock(return_value='hostname')), \
                patch('socket.getfqdn', MagicMock(return_value='hostname')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'hostname', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['1.2.3.4', '1.2.3.4', '1.2.3.4'])):
            self.assertEqual(network._generate_minion_id(), ['hostname', '1.2.3.4'])

    def test_generate_minion_id_platform_used(self):
        '''
        Test if platform.node is used for the first occurrence.
        The platform.node is most common hostname resolver before anything else.

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='very.long.and.complex.domain.name')), \
                patch('socket.gethostname', MagicMock(return_value='hostname')), \
                patch('socket.getfqdn', MagicMock(return_value='')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'hostname', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['1.2.3.4', '1.2.3.4', '1.2.3.4'])):
            self.assertEqual(network.generate_minion_id(), 'very.long.and.complex.domain.name')

    def test_generate_minion_id_platform_localhost_filtered(self):
        '''
        Test if localhost is filtered from the first occurrence.

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='localhost')), \
                patch('socket.gethostname', MagicMock(return_value='pick.me')), \
                patch('socket.getfqdn', MagicMock(return_value='hostname.domainname.blank')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'hostname', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['1.2.3.4', '1.2.3.4', '1.2.3.4'])):
            self.assertEqual(network.generate_minion_id(), 'hostname.domainname.blank')

    def test_generate_minion_id_platform_localhost_filtered_all(self):
        '''
        Test if any of the localhost is filtered from everywhere.

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='localhost')), \
                patch('socket.gethostname', MagicMock(return_value='ip6-loopback')), \
                patch('socket.getfqdn', MagicMock(return_value='ip6-localhost')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'localhost', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['127.0.0.1', '::1', 'fe00::0', 'fe02::1', '1.2.3.4'])):
            self.assertEqual(network.generate_minion_id(), '1.2.3.4')

    def test_generate_minion_id_platform_localhost_only(self):
        '''
        Test if there is no other choice but localhost.

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='localhost')), \
                patch('socket.gethostname', MagicMock(return_value='ip6-loopback')), \
                patch('socket.getfqdn', MagicMock(return_value='ip6-localhost')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'localhost', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['127.0.0.1', '::1', 'fe00::0', 'fe02::1'])):
            self.assertEqual(network.generate_minion_id(), 'localhost')

    def test_generate_minion_id_platform_fqdn(self):
        '''
        Test if fqdn is picked up.

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='localhost')), \
                patch('socket.gethostname', MagicMock(return_value='ip6-loopback')), \
                patch('socket.getfqdn', MagicMock(return_value='pick.me')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'localhost', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['127.0.0.1', '::1', 'fe00::0', 'fe02::1'])):
            self.assertEqual(network.generate_minion_id(), 'pick.me')

    def test_generate_minion_id_platform_localhost_addrinfo(self):
        '''
        Test if addinfo is picked up.

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='localhost')), \
                patch('socket.gethostname', MagicMock(return_value='ip6-loopback')), \
                patch('socket.getfqdn', MagicMock(return_value='ip6-localhost')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'pick.me', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['127.0.0.1', '::1', 'fe00::0', 'fe02::1'])):
            self.assertEqual(network.generate_minion_id(), 'pick.me')

    def test_generate_minion_id_platform_ip_addr_only(self):
        '''
        Test if IP address is the only what is used as a Minion ID in case no DNS name.

        :return:
        '''
        with patch('platform.node', MagicMock(return_value='localhost')), \
                patch('socket.gethostname', MagicMock(return_value='ip6-loopback')), \
                patch('socket.getfqdn', MagicMock(return_value='ip6-localhost')), \
                patch('socket.getaddrinfo', MagicMock(return_value=[(2, 3, 0, 'localhost', ('127.0.1.1', 0))])), \
                patch('salt.utils.files.fopen', mock_open()), \
                patch('salt.utils.network.ip_addrs', MagicMock(return_value=['127.0.0.1', '::1', 'fe00::0', 'fe02::1', '1.2.3.4'])):
            self.assertEqual(network.generate_minion_id(), '1.2.3.4')

    def test_gen_mac(self):
        with patch('random.randint', return_value=1) as random_mock:
            self.assertEqual(random_mock.return_value, 1)
            ret = network.gen_mac('00:16:3E')
            expected_mac = '00:16:3E:01:01:01'
            self.assertEqual(ret, expected_mac)

    def test_mac_str_to_bytes(self):
        self.assertRaises(ValueError, network.mac_str_to_bytes, '31337')
        self.assertRaises(ValueError, network.mac_str_to_bytes, '0001020304056')
        self.assertRaises(ValueError, network.mac_str_to_bytes, '00:01:02:03:04:056')
        self.assertRaises(ValueError, network.mac_str_to_bytes, 'a0:b0:c0:d0:e0:fg')
        self.assertEqual(b'\x10\x08\x06\x04\x02\x00', network.mac_str_to_bytes('100806040200'))
        self.assertEqual(b'\xf8\xe7\xd6\xc5\xb4\xa3', network.mac_str_to_bytes('f8e7d6c5b4a3'))

    def test_generate_minion_id_with_long_hostname(self):
        '''
        Validate the fix for:

        https://github.com/saltstack/salt/issues/51160
        '''
        long_name = 'localhost-abcdefghijklmnopqrstuvwxyz-abcdefghijklmnopqrstuvwxyz'
        with patch('socket.gethostname', MagicMock(return_value=long_name)):
            # An exception is raised if unicode is passed to socket.getfqdn
            minion_id = network.generate_minion_id()
        assert minion_id != '', minion_id



# -*- coding: utf-8 -*-
'''
unit tests for the grains state
'''

from __future__ import absolute_import, print_function, unicode_literals

# Import Python libs
import os
import contextlib

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.paths import TMP
from tests.support.unit import TestCase, skipIf
from tests.support.mock import NO_MOCK, NO_MOCK_REASON, MagicMock, patch

# Import salt libs
import salt.utils.files
import salt.utils.stringutils
import salt.utils.yaml
import salt.modules.grains as grainsmod
import salt.states.grains as grains
from salt.ext import six


@skipIf(NO_MOCK, NO_MOCK_REASON)
class GrainsTestCase(TestCase, LoaderModuleMockMixin):

    def setup_loader_modules(self):
        grains_test_dir = '__salt_test_state_grains'
        if not os.path.exists(os.path.join(TMP, grains_test_dir)):
            os.makedirs(os.path.join(TMP, grains_test_dir))
        loader_globals = {
            '__opts__': {
                'test': False,
                'conf_file': os.path.join(TMP, grains_test_dir, 'minion'),
                'cachedir':  os.path.join(TMP, grains_test_dir),
                'local': True,
            },
            '__salt__': {
                'cmd.run_all': MagicMock(return_value={
                    'pid': 5,
                    'retcode': 0,
                    'stderr': '',
                    'stdout': ''}),
                'grains.get':    grainsmod.get,
                'grains.set':    grainsmod.set,
                'grains.setval': grainsmod.setval,
                'grains.delval': grainsmod.delval,
                'grains.append': grainsmod.append,
                'grains.remove': grainsmod.remove,
                'saltutil.sync_grains': MagicMock()
            }
        }
        return {grains: loader_globals, grainsmod: loader_globals}

    def assertGrainFileContent(self, grains_string):
        if os.path.isdir(grains.__opts__['conf_file']):
            grains_file = os.path.join(
                grains.__opts__['conf_file'],
                'grains')
        else:
            grains_file = os.path.join(
                os.path.dirname(grains.__opts__['conf_file']),
                'grains')
        with salt.utils.files.fopen(grains_file, "r") as grf:
            grains_data = salt.utils.stringutils.to_unicode(grf.read())
        self.assertMultiLineEqual(grains_string, grains_data)

    @contextlib.contextmanager
    def setGrains(self, grains_data):
        with patch.dict(grains.__grains__, grains_data):
            with patch.dict(grainsmod.__grains__, grains_data):
                if os.path.isdir(grains.__opts__['conf_file']):
                    grains_file = os.path.join(
                        grains.__opts__['conf_file'],
                        'grains')
                else:
                    grains_file = os.path.join(
                        os.path.dirname(grains.__opts__['conf_file']), 'grains')
                with salt.utils.files.fopen(grains_file, "w+") as grf:
                    salt.utils.yaml.safe_dump(grains_data, grf, default_flow_style=False)
                yield

    # 'exists' function tests: 2

    def test_exists_missing(self):
        with self.setGrains({'a': 'aval'}):
            ret = grains.exists(name='foo')
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'Grain does not exist')
            self.assertEqual(ret['changes'], {})

    def test_exists_found(self):
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Grain already set
            ret = grains.exists(name='foo')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain exists')
            self.assertEqual(ret['changes'], {})

    # 'present' function tests: 12

    def test_present_add(self):
        # Set a non existing grain
        with self.setGrains({'a': 'aval'}):
            ret = grains.present(name='foo', value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['changes'], {'foo': 'bar'})
            self.assertEqual(grains.__grains__, {'a': 'aval', 'foo': 'bar'})
            self.assertGrainFileContent("a: aval\nfoo: bar\n")

        # Set a non existing nested grain
        with self.setGrains({'a': 'aval'}):
            ret = grains.present(name='foo:is:nested', value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['changes'], {'foo': {'is': {'nested': 'bar'}}})
            self.assertEqual(grains.__grains__, {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})
            self.assertGrainFileContent("a: aval\n"
                                        "foo:\n"
                                        "  is:\n"
                                        "    nested: bar\n"
            )

        # Set a non existing nested dict grain
        with self.setGrains({'a': 'aval'}):
            ret = grains.present(
                name='foo:is:nested',
                value={'bar': 'is a dict'})
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['changes'], {'foo': {'is': {'nested': {'bar': 'is a dict'}}}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': {'bar': 'is a dict'}}}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  is:\n"
                                      + "    nested:\n"
                                      + "      bar: is a dict\n"
            )

    def test_present_add_key_to_existing(self):
        with self.setGrains({'a': 'aval', 'foo': {'k1': 'v1'}}):
            # Fails setting a grain to a dict
            ret = grains.present(
                name='foo:k2',
                value='v2')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Set grain foo:k2 to v2')
            self.assertEqual(ret['changes'], {'foo': {'k2': 'v2', 'k1': 'v1'}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'k1': 'v1', 'k2': 'v2'}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  k1: v1\n"
                                      + "  k2: v2\n"
            )

    def test_present_already_set(self):
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Grain already set
            ret = grains.present(
                name='foo',
                value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain is already set')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': 'bar'})

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}}):
            # Nested grain already set
            ret = grains.present(
                name='foo:is:nested',
                value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain is already set')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}}):
            # Nested dict grain already set
            ret = grains.present(
                name='foo:is',
                value={'nested': 'bar'})
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain is already set')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})

    def test_present_overwrite(self):
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Overwrite an existing grain
            ret = grains.present(
                name='foo',
                value='newbar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['changes'], {'foo': 'newbar'})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': 'newbar'})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo: newbar\n"
            )

        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Clear a grain (set to None)
            ret = grains.present(
                name='foo',
                value=None)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['changes'], {'foo': None})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': None})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo: null\n"
            )

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}}):
            # Overwrite an existing nested grain
            ret = grains.present(
                name='foo:is:nested',
                value='newbar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['changes'], {'foo': {'is': {'nested': 'newbar'}}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': 'newbar'}}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  is:\n"
                                      + "    nested: newbar\n"
            )

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}}):
            # Clear a nested grain (set to None)
            ret = grains.present(
                name='foo:is:nested',
                value=None)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['changes'], {'foo': {'is': {'nested': None}}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': None}}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  is:\n"
                                      + "    nested: null\n"
            )

    def test_present_fail_overwrite(self):
        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'val'}}}):
            # Overwrite an existing grain
            ret = grains.present(
                name='foo:is',
                value='newbar')
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['changes'], {})
            self.assertEqual(ret['comment'], 'The key \'foo:is\' exists but is a dict or a list. Use \'force=True\' to overwrite.')
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': 'val'}}})

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'val'}}}):
            # Clear a grain (set to None)
            ret = grains.present(
                name='foo:is',
                value=None)
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['changes'], {})
            self.assertEqual(ret['comment'], 'The key \'foo:is\' exists but is a dict or a list. Use \'force=True\' to overwrite.')
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': 'val'}}})

    def test_present_fails_to_set_dict_or_list(self):
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Fails to overwrite a grain to a list
            ret = grains.present(
                name='foo',
                value=['l1', 'l2'])
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'The key \'foo\' exists and the '
                                           + 'given value is a dict or a list. '
                                           + 'Use \'force=True\' to overwrite.')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': 'bar'})

        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Fails setting a grain to a dict
            ret = grains.present(
                name='foo',
                value={'k1': 'v1'})
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'The key \'foo\' exists and the given '
                                           + 'value is a dict or a list. Use '
                                           + '\'force=True\' to overwrite.')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': 'bar'})

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}}):
            # Fails to overwrite a nested grain to a list
            ret = grains.present(
                name='foo,is,nested',
                value=['l1', 'l2'],
                delimiter=',')
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['changes'], {})
            self.assertEqual(ret['comment'], 'The key \'foo:is:nested\' exists and the '
                                           + 'given value is a dict or a list. '
                                           + 'Use \'force=True\' to overwrite.')
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}}):
            # Fails setting a nested grain to a dict
            ret = grains.present(
                name='foo:is:nested',
                value={'k1': 'v1'})
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'The key \'foo:is:nested\' exists and the '
                                           + 'given value is a dict or a list. '
                                           + 'Use \'force=True\' to overwrite.')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': 'bar'}}})

    def test_present_fail_merge_dict(self):
        with self.setGrains({'a': 'aval', 'foo': {'k1': 'v1'}}):
            # Fails setting a grain to a dict
            ret = grains.present(
                name='foo',
                value={'k2': 'v2'})
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'The key \'foo\' exists but '
                                           + 'is a dict or a list. '
                                           + 'Use \'force=True\' to overwrite.')
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'k1': 'v1'}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  k1: v1\n"
            )

    def test_present_force_to_set_dict_or_list(self):
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Force to overwrite a grain to a list
            ret = grains.present(
                name='foo',
                value=['l1', 'l2'],
                force=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(
                ret['comment'],
                "Set grain foo to ['l1', 'l2']" if six.PY3
                    else "Set grain foo to [u'l1', u'l2']"
            )
            self.assertEqual(ret['changes'], {'foo': ['l1', 'l2']})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['l1', 'l2']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- l1\n"
                                      + "- l2\n"
            )

        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Force setting a grain to a dict
            ret = grains.present(
                name='foo',
                value={'k1': 'v1'},
                force=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(
                ret['comment'],
                "Set grain foo to {'k1': 'v1'}" if six.PY3
                    else "Set grain foo to {u'k1': u'v1'}"
            )
            self.assertEqual(ret['changes'], {'foo': {'k1': 'v1'}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'k1': 'v1'}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  k1: v1\n"
            )

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}}}):
            # Force to overwrite a nested grain to a list
            ret = grains.present(
                name='foo,is,nested',
                value=['l1', 'l2'],
                delimiter=',',
                force=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['changes'], {'foo': {'is': {'nested': ['l1', 'l2']}}})
            self.assertEqual(
                ret['comment'],
                "Set grain foo:is:nested to ['l1', 'l2']" if six.PY3
                    else "Set grain foo:is:nested to [u'l1', u'l2']"
            )
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': ['l1', 'l2']}}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  is:\n"
                                      + "    nested:\n"
                                      + "    - l1\n"
                                      + "    - l2\n"
            )

        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': 'bar'}, 'and': 'other'}}):
            # Force setting a nested grain to a dict
            ret = grains.present(
                name='foo:is:nested',
                value={'k1': 'v1'},
                force=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(
                ret['comment'],
                "Set grain foo:is:nested to {'k1': 'v1'}" if six.PY3
                    else "Set grain foo:is:nested to {u'k1': u'v1'}"
            )
            self.assertEqual(ret['changes'], {'foo': {'is': {'nested': {'k1': 'v1'}}, 'and': 'other'}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': {'k1': 'v1'}}, 'and': 'other'}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  and: other\n"
                                      + "  is:\n"
                                      + "    nested:\n"
                                      + "      k1: v1\n"
            )

    def test_present_fails_to_convert_value_to_key(self):
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            # Fails converting a value to a nested grain key
            ret = grains.present(
                name='foo:is:nested',
                value={'k1': 'v1'})
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'The key \'foo\' value is \'bar\', '
                                   + 'which is different from the provided '
                                   + 'key \'is\'. Use \'force=True\' to overwrite.')
            self.assertEqual(ret['changes'], {})

    def test_present_overwrite_test(self):
        with patch.dict(grains.__opts__, {'test': True}):
            with self.setGrains({'a': 'aval', 'foo': 'bar'}):
                # Overwrite an existing grain
                ret = grains.present(
                    name='foo',
                    value='newbar')
                self.assertEqual(ret['result'], None)
                self.assertEqual(ret['changes'], {'changed': {'foo': 'newbar'}})
                self.assertEqual(
                    grains.__grains__,
                    {'a': 'aval', 'foo': 'bar'})
                self.assertGrainFileContent("a: aval\n"
                                          + "foo: bar\n"
                )

    def test_present_convert_value_to_key(self):
        with self.setGrains({'a': 'aval', 'foo': 'is'}):
            # Converts a value to a nested grain key
            ret = grains.present(
                name='foo:is:nested',
                value={'k1': 'v1'})
            self.assertEqual(ret['result'], True)
            self.assertEqual(
                ret['comment'],
                "Set grain foo:is:nested to {'k1': 'v1'}" if six.PY3
                    else "Set grain foo:is:nested to {u'k1': u'v1'}"
            )
            self.assertEqual(ret['changes'], {'foo': {'is': {'nested': {'k1': 'v1'}}}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': {'k1': 'v1'}}}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  is:\n"
                                      + "    nested:\n"
                                      + "      k1: v1\n"
            )

        with self.setGrains({'a': 'aval', 'foo': ['one', 'is', 'correct']}):
            # Converts a list element to a nested grain key
            ret = grains.present(
                name='foo:is:nested',
                value={'k1': 'v1'})
            self.assertEqual(ret['result'], True)
            self.assertEqual(
                ret['comment'],
                "Set grain foo:is:nested to {'k1': 'v1'}" if six.PY3
                    else "Set grain foo:is:nested to {u'k1': u'v1'}"
            )
            self.assertEqual(ret['changes'], {'foo': ['one', {'is': {'nested': {'k1': 'v1'}}}, 'correct']})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['one', {'is': {'nested': {'k1': 'v1'}}}, 'correct']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- one\n"
                                      + "- is:\n"
                                      + "    nested:\n"
                                      + "      k1: v1\n"
                                      + "- correct\n"
            )

    def test_present_unknown_failure(self):
        with patch('salt.modules.grains.setval') as mocked_setval:
            mocked_setval.return_value = 'Failed to set grain foo'
            with self.setGrains({'a': 'aval', 'foo': 'bar'}):
                # Unknown reason failure
                ret = grains.present(
                    name='foo',
                    value='baz')
                self.assertEqual(ret['result'], False)
                self.assertEqual(ret['comment'], 'Failed to set grain foo')
                self.assertEqual(ret['changes'], {})
                self.assertEqual(
                    grains.__grains__,
                    {'a': 'aval', 'foo': 'bar'})
                self.assertGrainFileContent("a: aval\n"
                                          + "foo: bar\n"
                )

    # 'absent' function tests: 6

    def test_absent_already(self):
        # Unset a non existent grain
        with self.setGrains({'a': 'aval'}):
            ret = grains.absent(
                name='foo')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain foo does not exist')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval'})
            self.assertGrainFileContent("a: aval\n")

        # Unset a non existent nested grain
        with self.setGrains({'a': 'aval'}):
            ret = grains.absent(
                name='foo:is:nested')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain foo:is:nested does not exist')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval'})
            self.assertGrainFileContent("a: aval\n")

    def test_absent_unset(self):
        # Unset a grain
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            ret = grains.absent(
                name='foo')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value for grain foo was set to None')
            self.assertEqual(ret['changes'], {'grain': 'foo', 'value': None})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': None})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo: null\n"
            )

        # Unset grain when its value is False
        with self.setGrains({'a': 'aval', 'foo': False}):
            ret = grains.absent(
                name='foo')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value for grain foo was set to None')
            self.assertEqual(ret['changes'], {'grain': 'foo', 'value': None})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': None})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo: null\n"
            )

        # Unset a nested grain
        with self.setGrains({'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar'}}, 'correct']}):
            ret = grains.absent(
                name='foo,is,nested',
                delimiter=',')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value for grain foo:is:nested was set to None')
            self.assertEqual(ret['changes'], {'grain': 'foo:is:nested', 'value': None})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['order', {'is': {'nested': None}}, 'correct']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- order\n"
                                      + "- is:\n"
                                      + "    nested: null\n"
                                      + "- correct\n"
            )

        # Unset a nested value don't change anything
        with self.setGrains({'a': 'aval', 'foo': ['order', {'is': 'nested'}, 'correct']}):
            ret = grains.absent(
                name='foo:is:nested')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain foo:is:nested does not exist')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['order', {'is': 'nested'}, 'correct']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- order\n"
                                      + "- is: nested\n"
                                      + "- correct\n"
            )

    def test_absent_unset_test(self):
        with patch.dict(grains.__opts__, {'test': True}):
            with self.setGrains({'a': 'aval', 'foo': 'bar'}):
                # Overwrite an existing grain
                ret = grains.absent(name='foo')
                self.assertEqual(ret['result'], None)
                self.assertEqual(ret['changes'], {'grain': 'foo', 'value': None})
                self.assertEqual(
                    grains.__grains__,
                    {'a': 'aval', 'foo': 'bar'})
                self.assertGrainFileContent("a: aval\n"
                                          + "foo: bar\n"
                )

    def test_absent_fails_nested_complex_grain(self):
        # Unset a nested complex grain
        with self.setGrains({'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar'}}, 'correct']}):
            ret = grains.absent(
                name='foo:is')
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'The key \'foo:is\' exists but is a dict or a list. Use \'force=True\' to overwrite.')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar'}}, 'correct']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- order\n"
                                      + "- is:\n"
                                      + "    nested: bar\n"
                                      + "- correct\n"
            )

    def test_absent_force_nested_complex_grain(self):
        # Unset a nested complex grain
        with self.setGrains({'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar'}}, 'correct']}):
            ret = grains.absent(
                name='foo:is',
                force=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value for grain foo:is was set to None')
            self.assertEqual(ret['changes'], {'grain': 'foo:is', 'value': None})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['order', {'is': None}, 'correct']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- order\n"
                                      + "- is: null\n"
                                      + "- correct\n"
            )

    def test_absent_delete(self):
        # Delete a grain
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            ret = grains.absent(
                name='foo',
                destructive=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain foo was deleted')
            self.assertEqual(ret['changes'], {'deleted': 'foo'})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval'})
            self.assertGrainFileContent("a: aval\n")

        # Delete a previously unset grain
        with self.setGrains({'a': 'aval', 'foo': None}):
            ret = grains.absent(
                name='foo',
                destructive=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain foo was deleted')
            self.assertEqual(ret['changes'], {'deleted': 'foo'})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval'})
            self.assertGrainFileContent("a: aval\n")

        # Delete a nested grain
        with self.setGrains({'a': 'aval', 'foo': ['order', {'is': {'nested': 'bar', 'other': 'value'}}, 'correct']}):
            ret = grains.absent(
                name='foo:is:nested',
                destructive=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain foo:is:nested was deleted')
            self.assertEqual(ret['changes'], {'deleted': 'foo:is:nested'})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['order', {'is': {'other': 'value'}}, 'correct']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- order\n"
                                      + "- is:\n"
                                      + "    other: value\n"
                                      + "- correct\n"
            )

    # 'append' function tests: 6

    def test_append(self):
        # Append to an existing list
        with self.setGrains({'a': 'aval', 'foo': ['bar']}):
            ret = grains.append(
                name='foo',
                value='baz')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value baz was added to grain foo')
            self.assertEqual(ret['changes'], {'added': 'baz'})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['bar', 'baz']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- bar\n"
                                      + "- baz\n"
            )

    def test_append_nested(self):
        # Append to an existing nested list
        with self.setGrains({'a': 'aval', 'foo': {'list': ['bar']}}):
            ret = grains.append(
                name='foo,list',
                value='baz',
                delimiter=',')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value baz was added to grain foo:list')
            self.assertEqual(ret['changes'], {'added': 'baz'})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'list': ['bar', 'baz']}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  list:\n"
                                      + "  - bar\n"
                                      + "  - baz\n"
            )

    def test_append_already(self):
        # Append to an existing list
        with self.setGrains({'a': 'aval', 'foo': ['bar']}):
            ret = grains.append(
                name='foo',
                value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value bar is already in the list '
                                           + 'for grain foo')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['bar']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- bar\n"
            )

    def test_append_fails_not_a_list(self):
        # Fail to append to an existing grain, not a list
        with self.setGrains({'a': 'aval', 'foo': {'bar': 'val'}}):
            ret = grains.append(
                name='foo',
                value='baz')
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'Grain foo is not a valid list')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'bar': 'val'}})

    def test_append_convert_to_list(self):
        # Append to an existing grain, converting to a list
        with self.setGrains({'a': 'aval', 'foo': {'bar': 'val'}}):
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  bar: val\n"
            )
            ret = grains.append(
                name='foo',
                value='baz',
                convert=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value baz was added to grain foo')
            self.assertEqual(ret['changes'], {'added': 'baz'})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': [{'bar': 'val'}, 'baz']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- bar: val\n"
                                      + "- baz\n"
            )

        # Append to an existing grain, converting to a list a multi-value dict
        with self.setGrains({'a': 'aval', 'foo': {'bar': 'val', 'other': 'value'}}):
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  bar: val\n"
                                      + "  other: value\n"
            )
            ret = grains.append(
                name='foo',
                value='baz',
                convert=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value baz was added to grain foo')
            self.assertEqual(ret['changes'], {'added': 'baz'})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': [{'bar': 'val', 'other': 'value'}, 'baz']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- bar: val\n"
                                      + "  other: value\n"
                                      + "- baz\n"
            )

    def test_append_fails_inexistent(self):
        # Append to a non existing grain
        with self.setGrains({'a': 'aval'}):
            ret = grains.append(
                name='foo',
                value='bar')
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'Grain foo does not exist')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval'})

    def test_append_convert_to_list_empty(self):
        # Append to an existing list
        with self.setGrains({'foo': None}):
            ret = grains.append(name='foo',
                                value='baz',
                                convert=True)
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value baz was added to grain foo')
            self.assertEqual(ret['changes'], {'added': 'baz'})
            self.assertEqual(
                grains.__grains__,
                {'foo': ['baz']})
            self.assertGrainFileContent("foo:\n"
                                      + "- baz\n")

    # 'list_present' function tests: 7

    def test_list_present(self):
        with self.setGrains({'a': 'aval', 'foo': ['bar']}):
            ret = grains.list_present(
                name='foo',
                value='baz')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Append value baz to grain foo')
            self.assertEqual(ret['changes'], {'new': {'foo': ['bar', 'baz']}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['bar', 'baz']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- bar\n"
                                      + "- baz\n"
            )

    def test_list_present_nested(self):
        with self.setGrains({'a': 'aval', 'foo': {'is': {'nested': ['bar']}}}):
            ret = grains.list_present(
                name='foo,is,nested',
                value='baz',
                delimiter=',')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Append value baz to grain foo:is:nested')
            self.assertEqual(ret['changes'], {'new': {'foo': {'is': {'nested': ['bar', 'baz']}}}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': ['bar', 'baz']}}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  is:\n"
                                      + "    nested:\n"
                                      + "    - bar\n"
                                      + "    - baz\n"
            )

    def test_list_present_inexistent(self):
        with self.setGrains({'a': 'aval'}):
            ret = grains.list_present(
                name='foo',
                value='baz')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Append value baz to grain foo')
            self.assertEqual(ret['changes'], {'new': {'foo': ['baz']}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['baz']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- baz\n"
            )

    def test_list_present_inexistent_nested(self):
        with self.setGrains({'a': 'aval'}):
            ret = grains.list_present(
                name='foo:is:nested',
                value='baz')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Append value baz to grain foo:is:nested')
            self.assertEqual(ret['changes'], {'new': {'foo': {'is': {'nested': ['baz']}}}})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'is': {'nested': ['baz']}}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  is:\n"
                                      + "    nested:\n"
                                      + "    - baz\n"
            )

    def test_list_present_not_a_list(self):
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            ret = grains.list_present(
                name='foo',
                value='baz')
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'Grain foo is not a valid list')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': 'bar'})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo: bar\n"
            )

    def test_list_present_nested_already(self):
        with self.setGrains({'a': 'aval', 'b': {'foo': ['bar']}}):
            ret = grains.list_present(
                name='b:foo',
                value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value bar is already in grain b:foo')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'b': {'foo': ['bar']}})
            self.assertGrainFileContent("a: aval\n"
                                      + "b:\n"
                                      + "  foo:\n"
                                      + "  - bar\n"
            )

    def test_list_present_already(self):
        with self.setGrains({'a': 'aval', 'foo': ['bar']}):
            ret = grains.list_present(
                name='foo',
                value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value bar is already in grain foo')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['bar']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- bar\n"
            )

    def test_list_present_unknown_failure(self):
        with self.setGrains({'a': 'aval', 'foo': ['bar']}):
            # Unknown reason failure

            with patch.dict(grainsmod.__salt__, {'grains.append': MagicMock()}):
                ret = grains.list_present(
                    name='foo',
                    value='baz')
                self.assertEqual(ret['result'], False)
                self.assertEqual(ret['comment'], 'Failed append value baz to grain foo')
                self.assertEqual(ret['changes'], {})
                self.assertEqual(
                    grains.__grains__,
                    {'a': 'aval', 'foo': ['bar']})
                self.assertGrainFileContent("a: aval\n"
                                          + "foo:\n"
                                          + "- bar\n"
                )

    # 'list_absent' function tests: 6

    def test_list_absent(self):
        with self.setGrains({'a': 'aval', 'foo': ['bar']}):
            ret = grains.list_absent(
                name='foo',
                value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value bar was deleted from grain foo')
            self.assertEqual(ret['changes'], {'deleted': ['bar']})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': []})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo: []\n"
            )

    def test_list_absent_nested(self):
        with self.setGrains({'a': 'aval', 'foo': {'list': ['bar']}}):
            ret = grains.list_absent(
                name='foo:list',
                value='bar')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value bar was deleted from grain foo:list')
            self.assertEqual(ret['changes'], {'deleted': ['bar']})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': {'list': []}})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "  list: []\n"
            )

    def test_list_absent_inexistent(self):
        with self.setGrains({'a': 'aval'}):
            ret = grains.list_absent(
                name='foo',
                value='baz')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain foo does not exist')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval'})
            self.assertGrainFileContent("a: aval\n")

    def test_list_absent_inexistent_nested(self):
        with self.setGrains({'a': 'aval'}):
            ret = grains.list_absent(
                name='foo:list',
                value='baz')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Grain foo:list does not exist')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval'})
            self.assertGrainFileContent("a: aval\n")

    def test_list_absent_not_a_list(self):
        with self.setGrains({'a': 'aval', 'foo': 'bar'}):
            ret = grains.list_absent(
                name='foo',
                value='bar')
            self.assertEqual(ret['result'], False)
            self.assertEqual(ret['comment'], 'Grain foo is not a valid list')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': 'bar'})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo: bar\n"
            )

    def test_list_absent_already(self):
        with self.setGrains({'a': 'aval', 'foo': ['bar']}):
            ret = grains.list_absent(
                name='foo',
                value='baz')
            self.assertEqual(ret['result'], True)
            self.assertEqual(ret['comment'], 'Value baz is absent from grain foo')
            self.assertEqual(ret['changes'], {})
            self.assertEqual(
                grains.__grains__,
                {'a': 'aval', 'foo': ['bar']})
            self.assertGrainFileContent("a: aval\n"
                                      + "foo:\n"
                                      + "- bar\n"
            )



# -*- coding: utf-8 -*-

# Import pytohn libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import skipIf, TestCase
from tests.support.mock import NO_MOCK, NO_MOCK_REASON, MagicMock, patch

# Import Salt libs
import salt.modules.nexus as nexus


@skipIf(NO_MOCK, NO_MOCK_REASON)
class nexusTestCase(TestCase, LoaderModuleMockMixin):

    def setup_loader_modules(self):
        return {nexus: {}}

    def test_artifact_get_metadata(self):
        with patch('salt.modules.nexus._get_artifact_metadata_xml', MagicMock(return_value='''<?xml version="1.0" encoding="UTF-8"?>
<metadata>
  <groupId>com.company.sampleapp.web-module</groupId>
  <artifactId>web</artifactId>
  <versioning>
    <release>0.1.0</release>
    <versions>
      <version>0.0.1</version>
      <version>0.0.2</version>
      <version>0.0.3</version>
      <version>0.1.0</version>
    </versions>
    <lastUpdated>20171010143552</lastUpdated>
  </versioning>
</metadata>''')):
            metadata = nexus._get_artifact_metadata(nexus_url='http://nexus.example.com/repository',
                                                repository='libs-releases',
                                                group_id='com.company.sampleapp.web-module',
                                                artifact_id='web',
                                                headers={})
            self.assertEqual(metadata['latest_version'], '0.1.0')

    def test_snapshot_version_get_metadata(self):
        with patch('salt.modules.nexus._get_snapshot_version_metadata_xml', MagicMock(return_value='''<?xml version="1.0" encoding="UTF-8"?>
<metadata modelVersion="1.1.0">
  <groupId>com.company.sampleapp.web-module</groupId>
  <artifactId>web</artifactId>
  <version>0.0.2-SNAPSHOT</version>
  <versioning>
    <snapshot>
      <timestamp>20170920.212353</timestamp>
      <buildNumber>3</buildNumber>
    </snapshot>
    <lastUpdated>20171112171500</lastUpdated>
    <snapshotVersions>
      <snapshotVersion>
        <classifier>sans-externalized</classifier>
        <extension>jar</extension>
        <value>0.0.2-20170920.212353-3</value>
        <updated>20170920212353</updated>
      </snapshotVersion>
      <snapshotVersion>
        <classifier>dist</classifier>
        <extension>zip</extension>
        <value>0.0.2-20170920.212353-3</value>
        <updated>20170920212353</updated>
      </snapshotVersion>
    </snapshotVersions>
  </versioning>
</metadata>''')):
            metadata = nexus._get_snapshot_version_metadata(nexus_url='http://nexus.example.com/repository',
                                                                 repository='libs-releases',
                                                                 group_id='com.company.sampleapp.web-module',
                                                                 artifact_id='web',
                                                                 version='0.0.2-SNAPSHOT',
                                                                 headers={})
            self.assertEqual(metadata['snapshot_versions']['dist'], '0.0.2-20170920.212353-3')

    def test_artifact_metadata_url(self):
        metadata_url = nexus._get_artifact_metadata_url(nexus_url='http://nexus.example.com/repository',
                                                             repository='libs-releases',
                                                             group_id='com.company.sampleapp.web-module',
                                                             artifact_id='web')

        self.assertEqual(metadata_url, "http://nexus.example.com/repository/libs-releases/com/company/sampleapp/web-module/web/maven-metadata.xml")

    def test_snapshot_version_metadata_url(self):
        metadata_url = nexus._get_snapshot_version_metadata_url(nexus_url='http://nexus.example.com/repository',
                                                             repository='libs-snapshots',
                                                             group_id='com.company.sampleapp.web-module',
                                                             artifact_id='web',
                                                             version='0.0.2-SNAPSHOT')

        self.assertEqual(metadata_url, "http://nexus.example.com/repository/libs-snapshots/com/company/sampleapp/web-module/web/0.0.2-SNAPSHOT/maven-metadata.xml")

    def test_construct_url_for_released_version(self):
        artifact_url, file_name = nexus._get_release_url(repository='libs-releases',
                                      group_id='com.company.sampleapp.web-module',
                                      artifact_id='web',
                                      packaging='zip',
                                      nexus_url='http://nexus.example.com/repository',
                                      version='0.1.0')

        self.assertEqual(artifact_url, "http://nexus.example.com/repository/libs-releases/com/company/sampleapp/web-module/web/0.1.0/web-0.1.0.zip")
        self.assertEqual(file_name, "web-0.1.0.zip")

    def test_construct_url_for_snapshot_version(self):
        with patch('salt.modules.nexus._get_snapshot_version_metadata', MagicMock(return_value={'snapshot_versions': {'zip': '0.0.2-20170920.212353-3'}})):

            artifact_url, file_name = nexus._get_snapshot_url(nexus_url='http://nexus.example.com/repository',
                                                   repository='libs-snapshots',
                                                   group_id='com.company.sampleapp.web-module',
                                                   artifact_id='web',
                                                   version='0.2.0-SNAPSHOT',
                                                   packaging='zip',
                                                   headers={})

            self.assertEqual(artifact_url, "http://nexus.example.com/repository/libs-snapshots/com/company/sampleapp/web-module/web/0.2.0-SNAPSHOT/web-0.0.2-20170920.212353-3.zip")
            self.assertEqual(file_name, "web-0.0.2-20170920.212353-3.zip")



# -*- coding: utf-8 -*-
'''
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
'''
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import skipIf, TestCase
from tests.support.mock import (
    NO_MOCK,
    NO_MOCK_REASON,
    MagicMock,
    patch)

# Import Salt Libs
import salt.states.modjk_worker as modjk_worker


@skipIf(NO_MOCK, NO_MOCK_REASON)
class ModjkWorkerTestCase(TestCase, LoaderModuleMockMixin):
    '''
    Test cases for salt.states.modjk_worker
    '''
    def setup_loader_modules(self):
        return {modjk_worker: {}}
    # 'stop' function tests: 1

    def test_stop(self):
        '''
        Test to stop the named worker from the lbn load balancers
         at the targeted minions.
        '''
        name = "{{ grains['id'] }}"
        lbn = 'application'
        target = 'roles:balancer'

        ret = {'name': name,
               'result': False,
               'comment': '',
               'changes': {}}

        comt = 'no servers answered the published command modjk.worker_status'
        mock = MagicMock(return_value=False)
        with patch.dict(modjk_worker.__salt__, {'publish.publish': mock}):
            ret.update({'comment': comt})
            self.assertDictEqual(modjk_worker.stop(name, lbn, target), ret)

    # 'activate' function tests: 1

    def test_activate(self):
        '''
        Test to activate the named worker from the lbn load balancers
         at the targeted minions.
        '''
        name = "{{ grains['id'] }}"
        lbn = 'application'
        target = 'roles:balancer'

        ret = {'name': name,
               'result': False,
               'comment': '',
               'changes': {}}

        comt = 'no servers answered the published command modjk.worker_status'
        mock = MagicMock(return_value=False)
        with patch.dict(modjk_worker.__salt__, {'publish.publish': mock}):
            ret.update({'comment': comt})
            self.assertDictEqual(modjk_worker.activate(name, lbn, target), ret)

    # 'disable' function tests: 1

    def test_disable(self):
        '''
        Test to disable the named worker from the lbn load balancers
         at the targeted minions.
        '''
        name = "{{ grains['id'] }}"
        lbn = 'application'
        target = 'roles:balancer'

        ret = {'name': name,
               'result': False,
               'comment': '',
               'changes': {}}

        comt = 'no servers answered the published command modjk.worker_status'
        mock = MagicMock(return_value=False)
        with patch.dict(modjk_worker.__salt__, {'publish.publish': mock}):
            ret.update({'comment': comt})
            self.assertDictEqual(modjk_worker.disable(name, lbn, target), ret)
